"""
This module implements helper functions used in both Freysoldt and Kumagai
correction metodules 
"""

import math
import warnings

import numpy as np

norm = np.linalg.norm

from doped.pycdt.utils.units import ang_to_bohr, eV_to_k, invang_to_ev

warnings.warn(
    "Replacing PyCDT correction utils with use "
    "corresponding objects in pymatgen.analysis.defects.corrections\n"
    "Will remove all PyCDT utils with Version 2.5 of PyCDT.",
    DeprecationWarning,
)


def cleanlat(dat):
    """
    TODO: Identify better function name and input variable name
    Args:
        dat: array of lattice vectors

    Returns:
        (double) Lattice constants (in same units as lattice vectors)
    """
    return list(map(norm, dat))


warnings.warn(
    "Replacing PyCDT correction utils with use "
    "corresponding objects in pymatgen.analysis.defects.corrections\n"
    "Will remove all PyCDT utils with Version 2.5 of PyCDT.",
    DeprecationWarning,
)


def genrecip(a1, a2, a3, encut):
    """
    Args:
        a1, a2, a3: lattice vectors in bohr
        encut: energy cut off in eV
    Returns:
        reciprocal lattice vectors with energy less than encut
    """
    vol = np.dot(a1, np.cross(a2, a3))  # 1/bohr^3
    b1 = (2 * np.pi / vol) * np.cross(a2, a3)  # units 1/bohr
    b2 = (2 * np.pi / vol) * np.cross(a3, a1)
    b3 = (2 * np.pi / vol) * np.cross(a1, a2)

    # create list of recip space vectors that satisfy |i*b1+j*b2+k*b3|<=encut
    gcut = eV_to_k(encut)
    imax = int(math.ceil(gcut / min(map(norm, [b1, b2, b3]))))

    for i in range(-imax, imax + 1):
        for j in range(-imax, imax + 1):
            for k in range(-imax, imax + 1):
                vec = i * b1 + j * b2 + k * b3
                en = invang_to_ev * (((1.0 / ang_to_bohr) * norm(vec)) ** 2)
                if en <= encut and en != 0:
                    yield vec


warnings.warn(
    "Replacing PyCDT correction utils with use "
    "corresponding objects in pymatgen.analysis.defects.corrections\n"
    "Will remove all PyCDT utils with Version 2.5 of PyCDT.",
    DeprecationWarning,
)


def generate_reciprocal_vectors_squared(a1, a2, a3, encut):
    """
    Generate reciprocal vector magnitudes within the cutoff along the specied
    lattice vectors.
    Args:
        a1: Lattice vector a (in Bohrs)
        a2: Lattice vector b (in Bohrs)
        a3: Lattice vector c (in Bohrs)
        encut: Reciprocal vector energy cutoff

    Returns:
        [[g1^2], [g2^2], ...] Square of reciprocal vectors (1/Bohr)^2
        determined by a1, a2, a3 and whose magntidue is less than gcut^2.
    """
    vol = np.dot(a1, np.cross(a2, a3))
    b1 = (2 * np.pi / vol) * np.cross(a2, a3)
    b2 = (2 * np.pi / vol) * np.cross(a3, a1)
    b3 = (2 * np.pi / vol) * np.cross(a1, a2)

    # Max (i,j,k) that doesn't upset the condition |i*b1+j*b2+k*b3|<=gcut
    gcut = eV_to_k(encut)
    imax = int(math.ceil(gcut / min(norm(b1), norm(b2), norm(b3))))
    gcut2 = gcut * gcut

    for i in range(-imax, imax + 1):
        for j in range(-imax, imax + 1):
            for k in range(-imax, imax + 1):
                vec = i * b1 + j * b2 + k * b3
                vec2 = np.dot(vec, vec)
                if vec2 <= gcut2 and vec2 != 0.0:
                    yield vec2


warnings.warn(
    "Replacing PyCDT correction utils with use "
    "corresponding objects in pymatgen.analysis.defects.corrections\n"
    "Will remove all PyCDT utils with Version 2.5 of PyCDT.",
    DeprecationWarning,
)


def closestsites(struct_blk, struct_def, pos):
    """
    Returns closest site to the input position
    for both bulk and defect structures
    Args:
        struct_blk: Bulk structure
        struct_def: Defect structure
        pos: Position
    Return: (site object, dist, index)
    """
    blk_close_sites = struct_blk.get_sites_in_sphere(pos, 5, include_index=True)
    blk_close_sites.sort(key=lambda x: x[1])
    def_close_sites = struct_def.get_sites_in_sphere(pos, 5, include_index=True)
    def_close_sites.sort(key=lambda x: x[1])

    return blk_close_sites[0], def_close_sites[0]


warnings.warn(
    "Replacing PyCDT correction utils with use "
    "corresponding objects in pymatgen.analysis.defects.corrections\n"
    "Will remove all PyCDT utils with Version 2.5 of PyCDT.",
    DeprecationWarning,
)


def find_defect_pos(struct_blk, struct_def, defpos=None):
    """
    output cartesian coords of defect in bulk,defect cells.

    Args:
        struct_blk:
        struct_def:
        defpos: (if known) defect position as a pymatgen Site object within the bulk supercell
            if this is not given, then defect position can be determined
            (but if large relaxation occured, this process can sometimes be problematic...)

    If vacancy defectpos=None,
    if interstitial bulkpos=None,
    if antisite/sub then both defined
    """
    if len(struct_blk.sites) > len(struct_def.sites):
        type_def = "vacancy"
        if defpos:
            return defpos.coords, None
    elif len(struct_blk.sites) < len(struct_def.sites):
        type_def = "interstitial"
        if defpos:
            return None, defpos.coords
    else:
        type_def = "substitution"  # also corresponds to antisite
        if defpos:
            return defpos.coords, defpos.coords

    sitematching = []
    foundindex = []
    for site in struct_blk.sites:
        blksite, defsite = closestsites(struct_blk, struct_def, site.coords)
        if type_def == "interstitial":
            foundindex.append(defsite[-1])
        if blksite[0].specie.symbol != defsite[0].specie.symbol:
            if type_def == "vacancy":
                return blksite[0].coords, None
            elif type_def == "interstitial":
                return None, defsite[0].coords
            else:  # subs or antisite type
                return blksite[0].coords, defsite[0].coords
        sitematching.append([blksite[0], blksite[1], defsite[0], defsite[1]])

    if type_def == "vacancy":
        # in case site type is same for closest site to vacancy
        sitematching.sort(key=lambda x: x[3])
        vacant = sitematching[-1]
        return vacant[0].coords, None
    elif type_def == "interstitial":
        for i, site in enumerate(struct_def.sites):
            if i not in foundindex:
                return None, site.coords
        # just in case site type is same for closest site to interstit
        sitematching.sort(key=lambda x: x[1])
        interstit = sitematching[-1]
        return None, interstit[2].coords

    return None, None  # if you get here there is an error



# coding: utf-8

"""
Code to generate VASP defect calculation input files.
"""

import functools
import os
import warnings
from copy import deepcopy  # See https://stackoverflow.com/a/22341377/14020960 why
from typing import TYPE_CHECKING, Optional, Union

from monty.io import zopen
from monty.serialization import dumpfn, loadfn
from pymatgen.io.vasp.inputs import (
    BadIncarWarning,
    incar_params,
    UnknownPotcarWarning,
    Incar,
    Kpoints,
    Poscar,
)
from pymatgen.io.vasp.sets import DictSet

from doped.pycdt.utils.vasp import DefectRelaxSet, _check_psp_dir

if TYPE_CHECKING:
    import pymatgen.core.periodic_table
    import pymatgen.core.structure

MODULE_DIR = os.path.dirname(os.path.abspath(__file__))
default_potcar_dict = loadfn(os.path.join(MODULE_DIR, "PotcarSet.yaml"))
default_relax_set = loadfn(os.path.join(MODULE_DIR, "HSE06_RelaxSet.yaml"))
default_defect_set = loadfn(os.path.join(MODULE_DIR, "DefectSet.yaml"))
default_relax_set["INCAR"].update(default_defect_set["INCAR"])

# globally ignore these POTCAR warnings
warnings.filterwarnings("ignore", category=UnknownPotcarWarning)
warnings.filterwarnings("ignore", message="No POTCAR file with matching TITEL fields")
warnings.filterwarnings("ignore", message="Ignoring unknown variable type")
warnings.filterwarnings(
    "ignore", message="POTCAR data with symbol"
)  # Ignore POTCAR warnings because Pymatgen incorrectly detecting POTCAR types
# Ignore because comment after 'ALGO = Normal' causes this unnecessary warning:
warnings.filterwarnings("ignore", message="Hybrid functionals only support")

# until updated from pymatgen==2022.7.25 :
warnings.filterwarnings(
    "ignore", message="Using `tqdm.autonotebook.tqdm` in notebook mode"
)


# TODO: Updated naming convention, to match that implemented in `ShakeNBreak`.
def scaled_ediff(natoms):
    """
    Returns a scaled EDIFF value for VASP calculations, based on the number of atoms in the
    structure. EDIFF is set to 1e-5 per 50 atoms in the supercell, with a maximum EDIFF of 1e-4.
    """
    ediff = float(f"{((natoms/50)*1e-5):.1g}")
    return ediff if ediff <= 1e-4 else 1e-4


def _prepare_vasp_defect_inputs(defects: dict) -> dict:
    """
    Generates a dictionary of key information to setup VASP defect calculations
    Args:
        defects (dict):
            Dictionary of defect-object-dictionaries from the `defects`
            attribute of the `ChargedDefectsStructures` class (from `defectsmaker.py`).
    """
    defect_input_dict = {}
    comb_defs = functools.reduce(
        lambda x, y: x + y, [defects[key] for key in defects if key != "bulk"]
    )

    for defect in comb_defs:
        for charge in defect["charges"]:
            supercell = defect["supercell"]
            dict_transf = {
                "defect_type": defect["name"],
                "defect_site": defect["unique_site"],
                "defect_supercell_site": defect["bulk_supercell_site"],
                "defect_multiplicity": defect["site_multiplicity"],
                "charge": charge,
                "supercell": supercell["size"],
            }
            if "substitution_specie" in defect:
                dict_transf["substitution_specie"] = defect["substitution_specie"]

            frac_coords = defect["unique_site"].frac_coords
            approx_coords = f"~[{frac_coords[0]:.4f},{frac_coords[1]:.4f},{frac_coords[2]:.4f}]"
            # Note this gets truncated to 40 characters in the CONTCAR: (this should be less than
            # 40 chars in all cases):
            poscar_comment = f"{defect['name']} {approx_coords} {charge}"
            defect_species = defect["name"] + f"_{charge}"

            defect_input_dict[defect_species] = {
                "Defect Structure": defect["supercell"]["structure"],
                "POSCAR Comment": poscar_comment,
                "Transformation Dict": dict_transf,
            }

    return defect_input_dict


def _prepare_vasp_files(
    single_defect_dict: dict,
    output_dir: str = ".",
    subfolder: str = "vasp_gam",
    vasp_type: str = "gam",
    user_incar_settings: dict = None,
    user_kpoints_settings: Optional[Union[dict, Kpoints]] = None,
    user_potcar_functional="PBE_54",
    user_potcar_settings=None,
    unperturbed_poscar: bool = False,
    write_files: bool = True,
) -> DefectRelaxSet:
    """
    Prepare DefectRelaxSet object for VASP defect supercell input file generation
    """
    supercell = single_defect_dict["Defect Structure"]
    poscar_comment = (
        single_defect_dict["POSCAR Comment"]
        if "POSCAR Comment" in single_defect_dict
        else None
    )
    transf_dict = single_defect_dict["Transformation Dict"]

    if subfolder is None:
        vaspinputdir = output_dir  # output_dir here = {orig output_dir}/{defect_species}
    else:
        vaspinputdir = f"{output_dir}/{subfolder}"

    if not os.path.exists(vaspinputdir) and write_files:
        os.makedirs(vaspinputdir)

    if unperturbed_poscar and write_files:
        poscar = Poscar(supercell)
        if poscar_comment:
            poscar.comment = poscar_comment
        poscar.write_file(vaspinputdir + "/POSCAR")

    if user_potcar_functional is not None:
        potcars = _check_psp_dir()
        if not potcars:
            if vasp_type == "gam":
                warnings.warn(
                    "POTCAR directory not set up with pymatgen (see the doped homepage: "
                    "https://github.com/SMTG-UCL/doped for instructions on setting this up). "
                    "This is required to generate `POTCAR` and `INCAR` files (to set `NELECT` "
                    "and `NUPDOWN`), so only `POSCAR` files will be generated."
                )
            elif unperturbed_poscar:  # vasp_std, vasp_ncl
                warnings.warn(
                    "POTCAR directory not set up with pymatgen, so only unperturbed POSCAR files "
                    "will be generated (POTCARs also needed to determine appropriate NELECT "
                    "setting in INCAR files)"
                )
            else:
                raise ValueError(
                    "POTCAR directory not set up with pymatgen (see the doped homepage: "
                    "https://github.com/SMTG-UCL/doped for instructions on setting this up). "
                    "This is required to generate `POTCAR` and `INCAR` files (to set `NELECT` "
                    "and `NUPDOWN`), so no input files will be generated here."
                )
            return None

    relax_set = deepcopy(default_relax_set)
    potcar_dict = deepcopy(default_potcar_dict)
    if user_potcar_functional:
        potcar_dict["POTCAR_FUNCTIONAL"] = user_potcar_functional
    if user_potcar_settings:
        potcar_dict["POTCAR"].update(user_potcar_settings)
    relax_set.update(potcar_dict)

    if user_incar_settings:
        for k in user_incar_settings.keys():
            # check INCAR flags and warn if they don't exist (typos)
            if k not in incar_params.keys():
                # this code is taken from pymatgen.io.vasp.inputs
                warnings.warn(  # but only checking keys, not values so we can add comments etc
                    f"Cannot find {k} from your user_incar_settings in the list of INCAR flags",
                    BadIncarWarning,
                )
        relax_set["INCAR"].update(user_incar_settings)
        if "EDIFF" not in user_incar_settings:  # set scaled
            relax_set["INCAR"]["EDIFF"] = scaled_ediff(len(supercell))
        if (
            "EDIFFG" not in user_incar_settings
            and vasp_type == "ncl"
            and relax_set["INCAR"]["IBRION"] == -1
            and relax_set["INCAR"]["NSW"] == 0
        ):
            # default SOC = singlepoint calc, remove EDIFFG & POTIM from INCAR to avoid confusion:
            del relax_set["INCAR"]["EDIFFG"]
            del relax_set["INCAR"]["POTIM"]

    if user_kpoints_settings is None and vasp_type == "gam":
        user_kpoints_settings = Kpoints().from_dict(
            {
                "comment": "kpoints from doped.vasp_gam_files()",
                "generation_style": "Gamma",
            }
        )
    elif user_kpoints_settings is None:  # vasp_std, vasp_ncl
        # use default vasp_std/ncl KPOINTS, Gamma-centred 2 x 2 x 2 mesh
        user_kpoints_settings = Kpoints.from_dict(
            {
                "comment": "kpoints from doped",
                "generation_style": "Gamma",
                "kpoints": [[2, 2, 2]],
            }
        )

    defect_relax_set = DefectRelaxSet(
        supercell,
        config_dict=relax_set,
        user_kpoints_settings=user_kpoints_settings,  # accepts Kpoints obj, so we can set comment
        charge=transf_dict["charge"],
    )
    defect_relax_set.output_dir = vaspinputdir  # assign attribute to later use in write_input()

    return defect_relax_set


def vasp_gam_files(
    defect_dict: dict,
    output_dir: str = ".",
    subfolder: str = "vasp_gam",
    user_incar_settings: dict = None,
    user_potcar_functional="PBE_54",
    user_potcar_settings=None,
    write_files=True,
    **kwargs,  # to allow POTCAR testing on GH Actions
) -> [DefectRelaxSet,]:
    """
    Generates input files for VASP Gamma-point-only (`vasp_gam`) coarse defect supercell
    relaxations. See the `HSE06_RelaxSet.yaml` and `DefectSet.yaml` files in the `doped` folder
    for the default `INCAR` settings, and `PotcarSet.yaml` for the default `POTCAR` settings.

    Note that any changes to the default `INCAR`/`POTCAR` settings should be consistent with
    those used for competing phase (chemical potential) calculations.

    Args:
        defect_dict (dict):
            Dictionary of defects to generate VASP input files for. Should match the output
            format of the `defects` attribute of `ChargedDefectsStructures` (as shown in the
            `dope_workflow_example` notebook) – i.e. {"substitutions": [list of substitution
            defects], "interstitials":...}, or a single defect subdict of this – i.e. {"name":
            "vac_1_Se", "supercell":...}.
        output_dir (str):
            Folder in which to create the VASP defect calculation inputs folders. Default is the
            current directory.
        subfolder (str):
            Output folder structure is `<defect_species>/<subfolder>` where `subfolder` =
            'vasp_gam' by default. Setting `subfolder` to `None` will write the `vasp_gam` input
            files directly to the `<defect_species>` folder, with no subfolders created.
        user_incar_settings (dict):
            Dictionary of user INCAR settings (AEXX, NCORE etc.) to override default settings.
            Highly recommended to look at output INCARs or the `HSE06_RelaxSet.yaml` and
            `DefectSet.yaml` files in the `doped` folder, to see what the default INCAR settings
            are. Note that any flags that aren't numbers or True/False need to be input as
            strings with quotation marks (e.g. `{"ALGO": "All"}`).
            (default: None)
        user_potcar_functional (str): POTCAR functional to use (default = "PBE_54")
        user_potcar_settings (dict): Override the default POTCARs, e.g. {"Li": "Li_sv"}. See
            `doped/PotcarSet.yaml` for the default `POTCAR` set.
        write_files (bool): Whether to write the VASP input files to disk. Default is True.

    Returns:
        Dictionary of {defect_species: `DefectRelaxSet`} for each defect species in the input
        defect_dict. `DefectRelaxSet` is a subclass of `pymatgen`'s `DictSet` class,
        with `incar`, `poscar`, `kpoints` and `potcar` attributes, containing information on the
        generated files.
    """
    # check if input dict is a single defect subdict:
    if "name" in defect_dict and "supercell" in defect_dict:  # single defect subdict
        defect_dict = {"defects": [defect_dict]}
    defect_input_dict = _prepare_vasp_defect_inputs(defect_dict)
    defect_relax_set_dict = {}
    for defect_species, single_defect_dict in defect_input_dict.items():
        defect_relax_set = _prepare_vasp_files(
            single_defect_dict=single_defect_dict,
            output_dir=f"{output_dir}/{defect_species}",
            subfolder=subfolder,
            vasp_type="gam",
            user_incar_settings=user_incar_settings,
            user_kpoints_settings=None,  # defaults to Gamma-only when vasp_type="gam"
            user_potcar_functional=user_potcar_functional,
            user_potcar_settings=user_potcar_settings,
            unperturbed_poscar=True,  # write POSCAR for vasp_gam_files()
            write_files=write_files,
        )

        if write_files:
            defect_relax_set.write_input(
                defect_relax_set.output_dir,
                **kwargs,  # kwargs to allow POTCAR testing on GH Actions
            )  # writes POSCAR without comment
            poscar_comment = (
                single_defect_dict["POSCAR Comment"]
                if "POSCAR Comment" in single_defect_dict
                else None
            )
            if poscar_comment is not None:
                poscar = Poscar(single_defect_dict["Defect Structure"])
                poscar.comment = poscar_comment
                poscar.write_file(defect_relax_set.output_dir + "/POSCAR")

            dumpfn(single_defect_dict["Transformation Dict"],  # write transformation.json file
                   f"{defect_relax_set.output_dir}/transformation.json")

        defect_relax_set_dict[defect_species] = defect_relax_set

    return defect_relax_set_dict


def vasp_std_files(
    defect_dict: dict,
    output_dir: str = ".",
    subfolder: str = "vasp_std",
    user_incar_settings: dict = None,
    user_kpoints_settings: Optional[Union[dict, Kpoints]] = None,
    user_potcar_functional="PBE_54",
    user_potcar_settings=None,
    unperturbed_poscar: bool = False,
    write_files: bool = True,
) -> [DefectRelaxSet,]:
    """
    Generates INCAR, POTCAR and KPOINTS for `vasp_std` defect supercell relaxations. By default
    does not generate POSCAR (input structure) files, as these should be taken from `ShakeNBreak`
    calculations (via `snb-groundstate`) or `vasp_gam` calculations (using
    `vasp_input.vasp_gam_files()`, and `cp vasp_gam/CONTCAR vasp_std/POSCAR`). See the
    `HSE06_RelaxSet.yaml` and `DefectSet.yaml` files in the `doped` folder for the default
    `INCAR` settings, and `PotcarSet.yaml` for the default `POTCAR` settings.

    Note that any changes to the default `INCAR`/`POTCAR` settings should be consistent with
    those used for competing phase (chemical potential) calculations.

    Args:
        defect_dict (dict):
            Dictionary of defects to generate VASP input files for. Should match the output
            format of the `defects` attribute of `ChargedDefectsStructures` (as shown in the
            `dope_workflow_example` notebook) – i.e. {"substitutions": [list of substitution
            defects], "interstitials":...}, or a single defect subdict of this – i.e. {"name":
            "vac_1_Se", "supercell":...}.
        output_dir (str):
            Folder in which to create the VASP defect calculation inputs folders. Default is the
            current directory.
        subfolder (str):
            Output folder structure is `<defect_species>/<subfolder>` where `subfolder` =
            'vasp_std' by default. Setting `subfolder` to `None` will write the vasp_std input
            files directly to the `<defect_species>` folder, with no subfolders created.
        user_incar_settings (dict):
            Dictionary of user INCAR settings (AEXX, NCORE etc.) to override default settings.
            Highly recommended to look at output INCARs or doped.vasp_input
            source code, to see what the default INCAR settings are. Note that any flags that
            aren't numbers or True/False need to be input as strings with quotation marks
            (e.g. `{"ALGO": "All"}`).
            (default: None)
        user_kpoints_settings (dict or Kpoints):
            Dictionary of user KPOINTS settings (in pymatgen DictSet() format) e.g.,
            {"reciprocal_density": 1000}, or a Kpoints object. Default KPOINTS is a Gamma-centred
            2 x 2 x 2 mesh.
            (default: None)
        user_potcar_functional (str): POTCAR functional to use (default = "PBE_54")
        user_potcar_settings (dict): Override the default POTCARs, e.g. {"Li": "Li_sv"}. See
            `doped/PotcarSet.yaml` for the default `POTCAR` set.
        unperturbed_poscar (bool):
            If True, write the unperturbed defect POSCARs to the generated folders as well. Not
            recommended, as the recommended workflow is to initially perform vasp_gam
            ground-state structure searching using ShakeNBreak (see example notebook;
            https://shakenbreak.readthedocs.io), then continue the vasp_std relaxations from the
            'Groundstate' CONTCARs.
            (default: False)
        write_files (bool): Whether to write the VASP input files to disk. Default is True.

    Returns:
        Dictionary of {defect_species: `DefectRelaxSet`} for each defect species in the input
        defect_dict. `DefectRelaxSet` is a subclass of `pymatgen`'s `DictSet` class,
        with `incar`, `poscar` (unperturbed), `kpoints` and `potcar` attributes, containing
        information on the generated files.
    """
    vaspstdincardict = {
        "KPAR": 2,  # vasp_std calculations so multiple k-points, likely quicker with this
    }
    if user_incar_settings is not None:
        vaspstdincardict.update(user_incar_settings)

    # check if input dict is a single defect subdict:
    if "name" in defect_dict and "supercell" in defect_dict:  # single defect subdict
        defect_dict = {"defects": [defect_dict]}

    defect_input_dict = _prepare_vasp_defect_inputs(defect_dict)
    defect_relax_set_dict = {}
    for defect_species, single_defect_dict in defect_input_dict.items():
        defect_relax_set = _prepare_vasp_files(
            single_defect_dict=single_defect_dict,
            output_dir=f"{output_dir}/{defect_species}",
            subfolder=subfolder,
            vasp_type="std",
            user_incar_settings=vaspstdincardict,
            user_kpoints_settings=user_kpoints_settings,
            user_potcar_functional=user_potcar_functional,
            user_potcar_settings=user_potcar_settings,
            unperturbed_poscar=unperturbed_poscar,
            write_files=write_files,
        )

        if write_files:
            # then use `write_file()`s rather than `write_input()` to avoid writing POSCARs
            defect_relax_set.incar.write_file(defect_relax_set.output_dir + "/INCAR")
            defect_relax_set.kpoints.write_file(defect_relax_set.output_dir + "/KPOINTS")
            if user_potcar_functional is not None:  # for GH Actions testing
                defect_relax_set.potcar.write_file(defect_relax_set.output_dir + "/POTCAR")

            dumpfn(single_defect_dict["Transformation Dict"],  # write transformation.json file
                   f"{defect_relax_set.output_dir}/transformation.json")

        defect_relax_set_dict[defect_species] = defect_relax_set

    return defect_relax_set_dict


def vasp_ncl_files(
    defect_dict: dict,
    output_dir: str = ".",
    subfolder: str = "vasp_ncl",
    user_incar_settings: dict = None,
    user_kpoints_settings: Optional[Union[dict, Kpoints]] = None,
    user_potcar_functional="PBE_54",
    user_potcar_settings=None,
    unperturbed_poscar: bool = False,
    write_files: bool = True,
) -> [DefectRelaxSet,]:
    """
    Generates INCAR, POTCAR and KPOINTS for `vasp_ncl` (i.e. spin-orbit coupling (SOC)) defect
    supercell singlepoint calculations. By default does not generate POSCAR (input structure)
    files, as these should be taken from `vasp_std` relaxations (i.e.
    `cp vasp_std/CONTCAR vasp_ncl/POSCAR`), which themselves likely originated from `ShakeNBreak`
    calculations (via `snb-groundstate`). See the `HSE06_RelaxSet.yaml` and `DefectSet.yaml`
    files in the `doped` folder for the default `INCAR` settings, and `PotcarSet.yaml` for the
    default `POTCAR` settings.

    Note that any changes to the default `INCAR`/`POTCAR` settings should be consistent with
    those used for competing phase (chemical potential) calculations.

    Args:
        defect_dict (dict):
            Dictionary of defects to generate VASP input files for. Should match the output
            format of the `defects` attribute of `ChargedDefectsStructures` (as shown in the
            `dope_workflow_example` notebook) – i.e. {"substitutions": [list of substitution
            defects], "interstitials":...}, or a single defect subdict of this – i.e. {"name":
            "vac_1_Se", "supercell":...}.
        output_dir (str):
            Folder in which to create the VASP defect calculation inputs folders. Default is the
            current directory.
        subfolder (str):
            Output folder structure is `<defect_species>/<subfolder>` where `subfolder` =
            'vasp_ncl' by default. Setting `subfolder` to `None` will write the vasp_ncl input
            files directly to the `<defect_species>` folder, with no subfolders created.
        user_incar_settings (dict):
            Dictionary of user INCAR settings (AEXX, NCORE etc.) to override default settings.
            Highly recommended to look at output INCARs or doped.vasp_input
            source code, to see what the default INCAR settings are. Note that any flags that
            aren't numbers or True/False need to be input as strings with quotation marks
            (e.g. `{"ALGO": "All"}`).
            (default: None)
        user_kpoints_settings (dict or Kpoints):
            Dictionary of user KPOINTS settings (in pymatgen DictSet() format) e.g.,
            {"reciprocal_density": 1000}, or a Kpoints object. Default KPOINTS is a Gamma-centred
            2 x 2 x 2 mesh.
            (default: None)
        user_potcar_functional (str): POTCAR functional to use (default = "PBE_54")
        user_potcar_settings (dict): Override the default POTCARs, e.g. {"Li": "Li_sv"}. See
            `doped/PotcarSet.yaml` for the default `POTCAR` set.
        unperturbed_poscar (bool):
            If True, write the unperturbed defect POSCARs to the generated folders as well. Not
            recommended, as the recommended workflow is to initially perform vasp_gam
            ground-state structure searching using ShakeNBreak (see example notebook;
            https://shakenbreak.readthedocs.io), then continue the vasp_std relaxations from the
            'Groundstate' CONTCARs, before doing final vasp_ncl singleshot calculations if SOC is
            important.
            (default: False)
        write_files (bool): Whether to write the VASP input files to disk. Default is True.

    Returns:
        Dictionary of {defect_species: `DefectRelaxSet`} for each defect species in the input
        defect_dict. `DefectRelaxSet` is a subclass of `pymatgen`'s `DictSet` class,
        with `incar`, `poscar` (unperturbed), `kpoints` and `potcar` attributes, containing
        information on the generated files.
    """
    vaspnclincardict = {
        "EDIFF": 1e-06,  # tight EDIFF for final energy and converged DOS",
        "KPAR": 2,  # vasp_ncl calculations so likely multiple k-points, likely quicker with this
        "LSORBIT": True,
        "NSW": 0,  # no ionic relaxation"
        "IBRION": -1,  # no ionic relaxation"
    }
    if user_incar_settings is not None:
        vaspnclincardict.update(user_incar_settings)

    # check if input dict is a single defect subdict:
    if "name" in defect_dict and "supercell" in defect_dict:  # single defect subdict
        defect_dict = {"defects": [defect_dict]}

    defect_input_dict = _prepare_vasp_defect_inputs(defect_dict)
    defect_relax_set_dict = {}
    for defect_species, single_defect_dict in defect_input_dict.items():
        defect_relax_set = _prepare_vasp_files(
            single_defect_dict=single_defect_dict,
            output_dir=f"{output_dir}/{defect_species}",
            subfolder=subfolder,
            vasp_type="ncl",
            user_incar_settings=vaspnclincardict,
            user_kpoints_settings=user_kpoints_settings,
            user_potcar_functional=user_potcar_functional,
            user_potcar_settings=user_potcar_settings,
            unperturbed_poscar=unperturbed_poscar,
            write_files=write_files,
        )

        if write_files:  # write INCAR, KPOINTS and POTCAR files
            # then use `write_file()`s rather than `write_input()` to avoid writing POSCARs
            defect_relax_set.incar.write_file(defect_relax_set.output_dir + "/INCAR")
            defect_relax_set.kpoints.write_file(defect_relax_set.output_dir + "/KPOINTS")
            if user_potcar_functional is not None:  # for GH Actions testing
                defect_relax_set.potcar.write_file(defect_relax_set.output_dir + "/POTCAR")

            dumpfn(single_defect_dict["Transformation Dict"],  # write transformation.json file
                   f"{defect_relax_set.output_dir}/transformation.json")

        defect_relax_set_dict[defect_species] = defect_relax_set

    return defect_relax_set_dict


# TODO: Remove these functions once confirming all functionality is in `chemical_potentials.py`;
# need `vasp_ncl_chempot` generation, `vaspup2.0` `input` folder with `CONFIG` generation as an
# option, improve chemical_potentials docstrings (i.e. mention defaults, note in notebooks if changing
# `INCAR`/`POTCAR` settings for competing phase production calcs, should also do with defect
# supercell calcs (and note this in vasp_input as well)), ensure consistent INCAR tags in defect
# supercell defaults and competing phase defaults, point too DefectSet in docstrings for defaults
# (noting the other INCAR tags that are changed)
def _vasp_converge_files(
    structure: "pymatgen.core.Structure",
    input_dir: str = None,
    incar_settings: dict = None,
    potcar_settings: dict = None,
    config: str = None,
) -> None:
    """
    Generates input files for single-shot GGA convergence test calculations.
    Automatically sets ISMEAR (in INCAR) to 2 (if metallic) or 0 if not.
    Recommended to use with vaspup2.0
    Args:
        structure (Structure object):
            Structure to create input files for.
        input_dir (str):
            Folder in which to create 'input' folder with VASP input files.
            (default: None)
        incar_settings (dict):
            Dictionary of user INCAR settings (AEXX, NCORE etc.) to override default settings.
            Highly recommended to look at output INCARs or doped.vasp_input
            source code, to see what the default INCAR settings are. Note that any flags that
            aren't numbers or True/False need to be input as strings with quotation marks
            (e.g. `{"ALGO": "All"}`).
            (default: None)
        config (str):
            CONFIG file string. If provided, will also write the CONFIG file (to automate
            convergence tests with vaspup2.0) to each 'input' directory.
            (default: None)
        potcar_settings (dict):
            Dictionary of user POTCAR settings to override default settings.
            Highly recommended to look at `default_potcar_dict` from doped.vasp_input to see what
            the (Pymatgen) syntax and doped default settings are.
            (default: None)
    """

    # Variable parameters first
    vaspconvergeincardict = {
        "# May need to change ISMEAR, NCORE, KPAR, AEXX, ENCUT, NUPDOWN, "
        + "ISPIN": "variable parameters",
        "NUPDOWN": "0 # But could be >0 if magnetic behaviour present",
        "NCORE": 12,
        "#KPAR": 1,
        "ENCUT": 400,
        "ISMEAR": "0 # Non-metal, use Gaussian smearing",
        "ISPIN": "1 # Change to 2 if spin polarisation or magnetic behaviour present",
        "GGA": "PS",  # PBEsol
        "ALGO": "Normal # Change to All if ZHEGV, FEXCP/F or ZBRENT errors encountered",
        "EDIFF": 1e-06,
        "EDIFFG": -0.01,
        "IBRION": -1,
        "ISIF": 3,
        "LASPH": True,
        "LORBIT": 14,
        "LREAL": False,
        "LWAVE": "False # Save filespace, shouldn't need WAVECAR from convergence tests",
        "NEDOS": 2000,
        "NELM": 100,
        "NSW": 0,
        "PREC": "Accurate",
        "SIGMA": 0.2,
    }
    if all(is_metal(element) for element in structure.composition.elements):
        vaspconvergeincardict[
            "ISMEAR"
        ] = "2 # Metal, use Methfessel-Paxton smearing scheme"
    if incar_settings:
        for (
            k
        ) in incar_settings.keys():  # check INCAR flags and warn if they don't exist (
            # typos)
            if (
                k not in incar_params.keys()
            ):  # this code is taken from pymatgen.io.vasp.inputs
                warnings.warn(  # but only checking keys, not values so we can add comments etc
                    "Cannot find %s from your user_incar_settings in the list of INCAR flags"
                    % (k),
                    BadIncarWarning,
                )
        vaspconvergeincardict.update(incar_settings)

    # Directory
    vaspconvergeinputdir = input_dir + "/input/" if input_dir else "VASP_Files/input/"
    if not os.path.exists(vaspconvergeinputdir):
        os.makedirs(vaspconvergeinputdir)

    # POTCAR
    potcar_dict = deepcopy(default_potcar_dict)
    if potcar_settings:
        if "POTCAR_FUNCTIONAL" in potcar_settings.keys():
            potcar_dict["POTCAR_FUNCTIONAL"] = potcar_settings["POTCAR_FUNCTIONAL"]
        if "POTCAR" in potcar_settings.keys():
            potcar_dict["POTCAR"].update(potcar_settings.pop("POTCAR"))
    vaspconvergeinput = DictSet(structure, config_dict=potcar_dict)
    vaspconvergeinput.potcar.write_file(vaspconvergeinputdir + "POTCAR")

    vaspconvergekpts = Kpoints().from_dict(
        {"comment": "Kpoints from vasp_gam_files", "generation_style": "Gamma"}
    )
    vaspconvergeincar = Incar.from_dict(vaspconvergeincardict)
    vaspconvergeincar.write_file(vaspconvergeinputdir + "INCAR")

    vaspconvergeposcar = Poscar(structure)
    vaspconvergeposcar.write_file(vaspconvergeinputdir + "POSCAR")

    vaspconvergekpts.write_file(vaspconvergeinputdir + "KPOINTS")
    # generate CONFIG file
    if config:
        with open(vaspconvergeinputdir + "CONFIG", "w+") as config_file:
            config_file.write(config)
        with open(vaspconvergeinputdir + "CONFIG", "a") as config_file:
            config_file.write(f"""\nname="{input_dir[13:]}" # input_dir""")


# Input files for vasp_std


def _vasp_std_chempot(
    structure: "pymatgen.core.Structure",
    input_dir: str = None,
    incar_settings: dict = None,
    kpoints_settings: dict = None,
    potcar_settings: dict = None,
) -> None:
    """
    Generates POSCAR, INCAR, POTCAR and KPOINTS for vasp_std chemical potentials relaxation.:
    Args:
        structure (Structure object):
            Structure to create input files for.
        input_dir (str):
            Folder in which to create vasp_std calculation inputs folder
            (default: None)
        incar_settings (dict):
            Dictionary of user INCAR settings (AEXX, NCORE etc.) to override default settings.
            Highly recommended to look at output INCARs or doped.vasp_input
            source code, to see what the default INCAR settings are. Note that any flags that
            aren't numbers or True/False need to be input as strings with quotation marks
            (e.g. `{"ALGO": "All"}`).
            (default: None)
        kpoints_settings (dict):
            Dictionary of user KPOINTS settings (in pymatgen Kpoints.from_dict() format). Common
            options would be "generation_style": "Monkhorst" (rather than "Gamma"),
            and/or "kpoints": [[3, 3, 1]] etc.
            Default KPOINTS is Gamma-centred 2 x 2 x 2 mesh.
            (default: None)
        potcar_settings (dict):
            Dictionary of user POTCAR settings to override default settings.
            Highly recommended to look at `default_potcar_dict` from doped.vasp_input to see what
            the (Pymatgen) syntax and doped default settings are.
            (default: None)
    """
    # INCAR Parameters
    vaspstdincardict = {
        "# May need to change NCORE, KPAR, ENCUT"
        + "ISPIN, POTIM": "variable parameters",
        "NCORE": 12,
        "KPAR": 2,
        "AEXX": 0.25,
        "ENCUT": 400,
        "POTIM": 0.2,
        "LSUBROT": "False # Change to True if relaxation poorly convergent",
        "ICORELEVEL": "0 # Needed if using the Kumagai-Oba (eFNV) anisotropic charge correction",
        "ALGO": "Normal # Change to All if ZHEGV, FEXCP/F or ZBRENT errors encountered",
        "EDIFF": 1e-06,  # May need to reduce for tricky relaxations",
        "EDIFFG": -0.01,
        "HFSCREEN": 0.2,  # assuming HSE06
        "IBRION": "1 # May need to change to 2 for difficult/poorly-convergent relaxations",
        "ISIF": 3,
        "ISMEAR": 0,
        "LASPH": True,
        "LHFCALC": True,
        "LORBIT": 14,
        "LREAL": False,
        "LVHAR": "True # Needed if using the Freysoldt (FNV) charge correction scheme",
        "LWAVE": True,
        "NEDOS": 2000,
        "NELM": 100,
        "NSW": 200,
        "PREC": "Accurate",
        "PRECFOCK": "Fast",
        "SIGMA": 0.05,
    }

    # Directory
    vaspstdinputdir = input_dir + "/vasp_std/" if input_dir else "VASP_Files/vasp_std/"
    if not os.path.exists(vaspstdinputdir):
        os.makedirs(vaspstdinputdir)

    # POTCAR
    potcar_dict = default_potcar_dict
    if potcar_settings:
        if "POTCAR_FUNCTIONAL" in potcar_settings.keys():
            potcar_dict["POTCAR_FUNCTIONAL"] = potcar_settings["POTCAR_FUNCTIONAL"]
        if "POTCAR" in potcar_settings.keys():
            potcar_dict["POTCAR"].update(potcar_settings.pop("POTCAR"))
    vaspstdinput = DictSet(structure, config_dict=potcar_dict)
    vaspstdinput.potcar.write_file(vaspstdinputdir + "POTCAR")

    if all(is_metal(element) for element in structure.composition.elements):
        vaspstdincardict["ISMEAR"] = "2 # Metal, use Methfessel-Paxton smearing scheme"
    if all(is_metal(element) for element in structure.composition.elements):
        vaspstdincardict["SIGMA"] = 0.02

    if incar_settings:
        for (
            k
        ) in (
            incar_settings.keys()
        ):  # check INCAR flags and warn if they don't exist (typos)
            if (
                k not in incar_params.keys()
            ):  # this code is taken from pymatgen.io.vasp.inputs
                warnings.warn(  # but only checking keys, not values so we can add comments etc
                    "Cannot find %s from your user_incar_settings in the list of INCAR flags"
                    % (k),
                    BadIncarWarning,
                )
        vaspstdincardict.update(incar_settings)

    # POSCAR
    vaspstdposcar = Poscar(structure)
    vaspstdposcar.write_file(vaspstdinputdir + "POSCAR")

    # KPOINTS
    vaspstdkpointsdict = {
        "comment": "Kpoints from doped.vasp_std_files",
        "generation_style": "Gamma",  # Set to Monkhorst for Monkhorst-Pack generation
        "kpoints": [[2, 2, 2]],
    }
    if kpoints_settings:
        vaspstdkpointsdict.update(kpoints_settings)
    vaspstdkpts = Kpoints.from_dict(vaspstdkpointsdict)
    vaspstdkpts.write_file(vaspstdinputdir + "KPOINTS")

    # INCAR
    vaspstdincar = Incar.from_dict(vaspstdincardict)
    with zopen(vaspstdinputdir + "INCAR", "wt") as incar_file:
        incar_file.write(vaspstdincar.get_string())


# Input files for vasp_ncl


def _vasp_ncl_chempot(
    structure: "pymatgen.core.Structure",
    input_dir: str = None,
    incar_settings: dict = None,
    kpoints_settings: dict = None,
    potcar_settings: dict = None,
) -> None:
    """
    Generates INCAR, POTCAR and KPOINTS for vasp_ncl chemical potentials relaxation.
    Take CONTCAR from vasp_std for POSCAR.:
    Args:
        structure (Structure object):
            Structure to create input files for.
        input_dir (str):
            Folder in which to create vasp_ncl calculation inputs folder
            (default: None)
        incar_settings (dict):
            Dictionary of user INCAR settings (AEXX, NCORE etc.) to override default settings.
            Highly recommended to look at output INCARs or doped.vasp_input
            source code, to see what the default INCAR settings are. Note that any flags that
            aren't numbers or True/False need to be input as strings with quotation marks
            (e.g. `{"ALGO": "All"}`).
            (default: None)
        kpoints_settings (dict):
            Dictionary of user KPOINTS settings (in pymatgen Kpoints.from_dict() format). Common
            options would be "generation_style": "Monkhorst" (rather than "Gamma"),
            and/or "kpoints": [[3, 3, 1]] etc.
            Default KPOINTS is Gamma-centred 2 x 2 x 2 mesh.
            (default: None)
        potcar_settings (dict):
            Dictionary of user POTCAR settings to override default settings.
            Highly recommended to look at `default_potcar_dict` from doped.vasp_input to see what
            the (Pymatgen) syntax and doped default settings are.
            (default: None)
    """
    # INCAR Parameters
    vaspnclincardict = {
        "# May need to change NELECT, NCORE, KPAR, AEXX, ENCUT, NUPDOWN": "variable parameters",
        "NCORE": 12,
        "KPAR": 2,
        "AEXX": 0.25,
        "ENCUT": 400,
        "ICORELEVEL": "0 # Needed if using the Kumagai-Oba (eFNV) anisotropic charge correction",
        "NSW": 0,
        "LSORBIT": True,
        "EDIFF": 1e-06,  # tight for final energy and converged DOS
        "EDIFFG": -0.01,
        "ALGO": "Normal # Change to All if ZHEGV, FEXCP/F or ZBRENT errors encountered",
        "HFSCREEN": 0.2,
        "IBRION": -1,
        "ISYM": 0,
        "ISMEAR": 0,
        "LASPH": True,
        "LHFCALC": True,
        "LORBIT": 14,
        "LREAL": False,
        "LVHAR": "True # Needed if using the Freysoldt (FNV) charge correction scheme",
        "LWAVE": True,
        "NEDOS": 2000,
        "NELM": 100,
        "PREC": "Accurate",
        "PRECFOCK": "Fast",
        "SIGMA": 0.05,
    }

    # Directory
    vaspnclinputdir = input_dir + "/vasp_ncl/" if input_dir else "VASP_Files/vasp_ncl/"
    if not os.path.exists(vaspnclinputdir):
        os.makedirs(vaspnclinputdir)

    # POTCAR
    potcar_dict = default_potcar_dict
    if potcar_settings:
        if "POTCAR_FUNCTIONAL" in potcar_settings.keys():
            potcar_dict["POTCAR_FUNCTIONAL"] = potcar_settings["POTCAR_FUNCTIONAL"]
        if "POTCAR" in potcar_settings.keys():
            potcar_dict["POTCAR"].update(potcar_settings.pop("POTCAR"))
    vaspnclinput = DictSet(structure, config_dict=potcar_dict)
    vaspnclinput.potcar.write_file(vaspnclinputdir + "POTCAR")

    if all(is_metal(element) for element in structure.composition.elements):
        vaspnclincardict["ISMEAR"] = "2 # Metal, use Methfessel-Paxton smearing scheme"
    if all(is_metal(element) for element in structure.composition.elements):
        vaspnclincardict["SIGMA"] = 0.02

    if incar_settings:
        for (
            k
        ) in (
            incar_settings.keys()
        ):  # check INCAR flags and warn if they don't exist (typos)
            if (
                k not in incar_params.keys()
            ):  # this code is taken from pymatgen.io.vasp.inputs
                warnings.warn(  # but only checking keys, not values so we can add comments etc
                    "Cannot find %s from your user_incar_settings in the list of INCAR flags"
                    % (k),
                    BadIncarWarning,
                )
        vaspnclincardict.update(incar_settings)

    # KPOINTS
    vaspnclkpointsdict = {
        "comment": "Kpoints from doped.vasp_ncl_files",
        "generation_style": "Gamma",  # Set to Monkhorst for Monkhorst-Pack generation
        "kpoints": [[2, 2, 2]],
    }
    if kpoints_settings:
        vaspnclkpointsdict.update(kpoints_settings)
    vaspnclkpts = Kpoints.from_dict(vaspnclkpointsdict)
    vaspnclkpts.write_file(vaspnclinputdir + "KPOINTS")

    # INCAR
    vaspnclincar = Incar.from_dict(vaspnclincardict)
    with zopen(vaspnclinputdir + "INCAR", "wt") as incar_file:
        incar_file.write(vaspnclincar.get_string())



# coding: utf-8

from __future__ import division

__status__ = "Development"

import unittest

import numpy as np
from pymatgen.analysis.defects.core import DefectEntry, Vacancy
from pymatgen.core.sites import PeriodicSite
from pymatgen.util.testing import PymatgenTest

from doped.pycdt.corrections.finite_size_charge_correction import (
    get_correction_freysoldt,
    get_correction_kumagai,
)


class FiniteSizeChargeCorrectionTest(PymatgenTest):
    """
    Test functions for getting freysoldt and kumagai corrections
    """

    def setUp(self):
        self.epsilon = 15.0

        struc = PymatgenTest.get_structure("VO2")
        struc.make_supercell(3)
        vac = Vacancy(struc, struc.sites[0], charge=-3)

        # load neccessary parameters for defect_entry to make use
        # of Freysoldt and Kumagai corrections
        p = {}
        ids = vac.generate_defect_structure(1)
        abc = struc.lattice.abc
        axisdata = [np.arange(0.0, lattval, 0.2) for lattval in abc]
        bldata = [
            np.array([1.0 for u in np.arange(0.0, lattval, 0.2)]) for lattval in abc
        ]
        dldata = [
            np.array(
                [
                    (-1 - np.cos(2 * np.pi * u / lattval))
                    for u in np.arange(0.0, lattval, 0.2)
                ]
            )
            for lattval in abc
        ]
        p.update(
            {
                "axis_grid": axisdata,
                "bulk_planar_averages": bldata,
                "defect_planar_averages": dldata,
                "initial_defect_structure": ids,
                "defect_frac_sc_coords": struc.sites[0].frac_coords,
                "bulk_sc_structure": struc,
            }
        )

        bulk_atomic_site_averages, defect_atomic_site_averages = [], []
        defect_site_with_sc_lattice = PeriodicSite(
            struc.sites[0].specie,
            struc.sites[0].coords,
            struc.lattice,
            coords_are_cartesian=True,
        )
        max_dist = 9.6
        pert_amnt = 1.0
        for site_ind, site in enumerate(struc.sites):
            if site.specie.symbol == "O":
                Oval = -30.6825
                bulk_atomic_site_averages.append(Oval)
                if site_ind:
                    dist_to_defect = site.distance_and_image(
                        defect_site_with_sc_lattice
                    )[0]
                    defect_site_val = (
                        Oval
                        - 0.3
                        + pert_amnt * ((max_dist - dist_to_defect) / max_dist) ** 2
                    )
                    defect_atomic_site_averages.append(defect_site_val)
            else:
                Vval = -51.6833
                bulk_atomic_site_averages.append(Vval)
                if site_ind:
                    dist_to_defect = site.distance_and_image(
                        defect_site_with_sc_lattice
                    )[0]
                    defect_site_val = (
                        Vval
                        - 0.3
                        + pert_amnt * ((max_dist - dist_to_defect) / max_dist) ** 2
                    )
                    defect_atomic_site_averages.append(defect_site_val)

        site_matching_indices = [
            [ind, ind - 1] for ind in range(len(struc.sites)) if ind != 0
        ]

        p.update(
            {
                "bulk_atomic_site_averages": bulk_atomic_site_averages,
                "defect_atomic_site_averages": defect_atomic_site_averages,
                "site_matching_indices": site_matching_indices,
            }
        )
        self.defect_entry = DefectEntry(vac, 0.0, parameters=p)

    def test_get_correction_freysoldt(self):
        freyout = get_correction_freysoldt(
            self.defect_entry, self.epsilon, title=None, partflag="All", axis=None
        )
        self.assertEqual(freyout, 5.445950368792991)

        freyout = get_correction_freysoldt(
            self.defect_entry, self.epsilon, title=None, partflag="AllSplit", axis=None
        )
        self.assertEqual(freyout[0], 0.975893)
        self.assertEqual(freyout[1], 4.4700573687929905)
        self.assertEqual(freyout[2], 5.445950368792991)

    def test_get_correction_kumagai(self):
        kumagaiout = get_correction_kumagai(
            self.defect_entry, self.epsilon, title=None, partflag="AllSplit"
        )
        self.assertEqual(kumagaiout[0], 0.9763991294314076)
        self.assertEqual(kumagaiout[1], 0.2579750033409367)
        self.assertEqual(kumagaiout[2], 1.2343741327723443)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8

from __future__ import division

__status__ = "Development"

import os
import unittest

import numpy as np
from pymatgen.core.structure import Structure

from doped.pycdt.corrections.utils import *

bs_path = os.path.abspath(
    os.path.join(__file__, "..", "..", "..", "..", "test_files", "POSCAR_Ga4As4")
)
ds_path = os.path.abspath(
    os.path.join(__file__, "..", "..", "..", "..", "test_files", "POSCAR_Ga3As4")
)


class StructureFunctionsTest(unittest.TestCase):
    def setUp(self):
        self.bs = Structure.from_file(bs_path)
        self.ds = Structure.from_file(ds_path)
        self.a = self.bs.lattice.matrix[0]
        self.b = self.bs.lattice.matrix[1]
        self.c = self.bs.lattice.matrix[2]

    def test_cleanlat(self):
        self.assertAlmostEqual(
            cleanlat(self.bs.lattice.matrix), [5.750183, 5.750183, 5.750183]
        )

    def test_genrecip(self):
        brecip = map(
            np.array,
            [
                [-1.0926931033637688, 0.0, 0.0],
                [0.0, -1.0926931033637688, 0.0],
                [0.0, 0.0, -1.0926931033637688],
                [0.0, 0.0, 1.0926931033637688],
                [0.0, 1.0926931033637688, 0.0],
                [1.0926931033637688, 0.0, 0.0],
            ],
        )
        recip = genrecip(self.a, self.b, self.c, 1.3)
        for vec in brecip:
            self.assertTrue(np.isclose(next(recip), vec).all())

    def test_generate_reciprocal_vectors_squared(self):
        brecip = [1.1939782181387439 for i in range(6)]
        self.assertAlmostEqual(
            list(generate_reciprocal_vectors_squared(self.a, self.b, self.c, 1.3)),
            brecip,
        )

    def test_closestsites(self):
        pos = [0.0000, 2.8751, 2.8751]
        bsite, dsite = closestsites(self.bs, self.ds, pos)
        self.assertAlmostEqual(list(bsite[0].coords), list(self.bs.sites[1].coords))
        self.assertAlmostEqual(list(dsite[0].coords), list(self.ds.sites[0].coords))

    def test_find_defect_pos(self):
        """
        TODO: might be good to include more defect poscars to test all
        types of defect finding.
        """
        bdefect, ddefect = find_defect_pos(self.bs, self.ds)
        self.assertAlmostEqual(list(bdefect), [0, 0, 0])
        self.assertIsNone(ddefect)


import unittest

if __name__ == "__main__":
    unittest.main()



# coding: utf-8
from __future__ import division

"""
Code to generate charged defects structure files.
"""


import abc
import re
from tabulate import tabulate

from monty.serialization import dumpfn
from pymatgen.analysis.defects.core import Interstitial
from pymatgen.analysis.defects.generators import (
    InterstitialGenerator,
    SimpleChargeGenerator,
    SubstitutionGenerator,
    VacancyGenerator,
    VoronoiInterstitialGenerator,
)
from pymatgen.analysis.local_env import ValenceIonicRadiusEvaluator as VIRE
from pymatgen.core.periodic_table import Element, Specie, get_el_sp
from pymatgen.core.structure import PeriodicSite, Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer


def get_optimized_sc_scale(inp_struct, final_site_no):
    """
    Get the optimal scaling to generate supercells with atoms less than
    the final_site_no.
    """

    if final_site_no < len(inp_struct.sites):
        final_site_no = len(inp_struct.sites)

    dictio = {}
    result = []
    for k1 in range(1, 6):
        for k2 in range(1, 6):
            for k3 in range(1, 6):
                struct = inp_struct.copy()
                struct.make_supercell([k1, k2, k3])
                if len(struct.sites) > final_site_no:
                    continue

                min_dist = 1000.0
                for a in range(-1, 2):
                    for b in range(-1, 2):
                        for c in range(-1, 2):
                            try:
                                distance = struct.get_distance(0, 0, (a, b, c))
                            except:
                                print(a, b, c)
                                raise
                            if distance < min_dist and distance > 0.00001:
                                min_dist = distance
                min_dist = round(min_dist, 3)
                if min_dist in dictio:
                    if dictio[min_dist]["num_sites"] > struct.num_sites:
                        dictio[min_dist]["num_sites"] = struct.num_sites
                        dictio[min_dist]["supercell"] = [k1, k2, k3]
                else:
                    dictio[min_dist] = {}
                    dictio[min_dist]["num_sites"] = struct.num_sites
                    dictio[min_dist]["supercell"] = [k1, k2, k3]
    min_dist = -1.0
    biggest = None
    for c in dictio:
        if c > min_dist:
            biggest = dictio[c]["supercell"]
            min_dist = c
    if biggest is None or min_dist < 0.0:
        raise RuntimeError("could not find any supercell scaling vector")
    return biggest


class DefectCharger:
    __metaclass__ = abc.ABCMeta
    """
    Abstract base class to define the properties of a defect charge generator
    """

    def __init__(self, structure):
        pass

    @abc.abstractmethod
    def get_charges(self, defect_type, site_specie=None, sub_specie=None):
        """
        Based on the type of defect, site and substitution (if any) species
        the defect charge states are generated.
        Args:
            defect_type (str): Options are vacancy, antisite, substitution,
                               and interstitial
            site_specie: Specie on the host lattice site
                         For interstitials, use this
            sub_specie: Specie that is replacing the site specie.
                        For antisites and substitution defects

        """
        raise NotImplementedError


class DefectChargerSemiconductor(DefectCharger):
    """
    Determine oxidation states from bond valence method (unless oxidation states specified)
    Use BVM or specified oxidation states for site oxidation preference
    Use possible oxidation state list for sub_site elements
    """

    def __init__(self, structure, min_max_oxi=None, oxi_states=None):
        """
        Charge assignment based on the oxidation states referenced from
        semiconductor database. Targetted materials are shallow and some
        wideband semiconductors. For these systems, antisites are common and
        their charge assignment for antisites follows vacancies
        Args: structure
            pymatgen structure object to determine the oxidation states
            min_max_oxi: any user specified min/max oxidation ranges for elements in structure
            oxi_states: any user specified oxidation states of elements in structure
        """
        min_max_oxi = min_max_oxi if min_max_oxi is not None else {}
        oxi_states = oxi_states if oxi_states is not None else {}

        struct_species = structure.types_of_specie
        if (len(struct_species) == 1) and struct_species[
            0
        ].symbol not in oxi_states.keys():
            oxi_states[struct_species[0].symbol] = 0
        else:
            vir = VIRE(structure)
            for elt, oxi in vir.valences.items():
                strip_key = "".join([s for s in elt if s.isalpha()])
                if strip_key not in oxi_states.keys():
                    oxi_states[strip_key] = oxi
        self.oxi_states = oxi_states

        self.min_max_oxi_bulk = [0, 0]
        for s in struct_species:
            if isinstance(s, Specie):
                el = s.element
            elif isinstance(s, Element):
                el = s
            else:
                continue
            if el.symbol not in min_max_oxi.keys():
                max_oxi = max(el.common_oxidation_states)
                min_oxi = min(el.common_oxidation_states)
                min_max_oxi[el.symbol] = [min_oxi, max_oxi]
            if min_max_oxi[el.symbol][0] < self.min_max_oxi_bulk[0]:
                self.min_max_oxi_bulk[0] = min_max_oxi[el.symbol][0]
            if min_max_oxi[el.symbol][1] > self.min_max_oxi_bulk[1]:
                self.min_max_oxi_bulk[1] = min_max_oxi[el.symbol][1]
        self.min_max_oxi = min_max_oxi

    def get_charges(self, defect_type, site_specie, sub_specie=None):
        """
        Based on the type of defect, site and substitution (if any) species
        the defect charge states are generated.
        Args:
            defect_type (str): Options are vacancy, antisite, substitution,
                               and interstitial
            site_specie (str): Specie on the host lattice site
                         Use this for interstitials as well
            sub_specie (str): Specie that is replacing the site specie.
                        At present used for substitution and antisite defects
        """
        if site_specie not in self.min_max_oxi.keys():
            max_oxi = max(Element(site_specie).common_oxidation_states)
            min_oxi = min(Element(site_specie).common_oxidation_states)
            self.min_max_oxi[site_specie] = [min_oxi, max_oxi]
        if sub_specie:
            if sub_specie not in self.min_max_oxi.keys():
                max_oxi = max(Element(sub_specie).common_oxidation_states)
                min_oxi = min(Element(sub_specie).common_oxidation_states)
                self.min_max_oxi[sub_specie] = [min_oxi, max_oxi]
        if defect_type == "vacancy":
            site_oxi = self.oxi_states[site_specie]
            if site_oxi:
                return list(range(-abs(site_oxi), abs(site_oxi) + 1))
            else:
                min_oxi = self.min_max_oxi[site_specie][0]
                max_oxi = self.min_max_oxi[site_specie][1]
                if abs(min_oxi) < abs(max_oxi):
                    return list(range(-abs(max_oxi), abs(max_oxi) + 1))
                else:
                    return list(range(-abs(min_oxi), abs(min_oxi) + 1))
        elif defect_type == "antisite":
            return list(range(self.min_max_oxi_bulk[0], self.min_max_oxi_bulk[1] - 1))
        elif defect_type == "substitution":
            oxi_sub_set = Element(sub_specie).common_oxidation_states
            oxi_site = self.oxi_states[site_specie]
            min_max_oxi_bulk_sub = [
                min(min(oxi_sub_set) - oxi_site, -1),
                max(max(oxi_sub_set) - oxi_site, 1),
            ]
            if (min_max_oxi_bulk_sub[1] - min_max_oxi_bulk_sub[0]) > 2:
                if min_max_oxi_bulk_sub[1] > 2:
                    return list(
                        range(min_max_oxi_bulk_sub[0], min_max_oxi_bulk_sub[1] - 2)
                    )  # if range exists, less likely to be a higher charge
                else:
                    return list(
                        range(min_max_oxi_bulk_sub[0], 2)
                    )  # extend to 2 if upper range does not already include this
            else:
                return list(
                    range(min_max_oxi_bulk_sub[0] - 1, min_max_oxi_bulk_sub[1] + 2)
                )  # likely upper bound is 0, so extend to 2
        elif defect_type == "interstitial":
            if self.min_max_oxi[site_specie][0] > 0:
                min_oxi = 0
            else:
                min_oxi = self.min_max_oxi[site_specie][0]
            if self.min_max_oxi[site_specie][1] < 0:
                max_oxi = 0
            else:
                max_oxi = self.min_max_oxi[site_specie][1]
            return list(range(min_oxi, max_oxi + 1))
        else:
            raise ValueError("Defect type not understood")


class DefectChargerInsulator(DefectCharger):
    """
    Conservative charge assignment based on the oxidation statess determined
    by bond valence. Targetted materials are wideband semiconductors and
    insulators. AxBy where A is cation and B is anion will have charge
    assignments {A: [0:y], B:[-x:0]}. For these systems, antisites typically
    have very high formation energies and are ignored.
    """

    def __init__(self, structure):
        """
        Conservative defect charge generator based on the oxidation statess
        determined by bond valence. Targetted materials are wideband
        semiconductors and insulators. AxBy where A is cation and B is
        anion will have charge assignments {A: [0:y], B:[-x:0]}. For these
        systems, antisites typically have very high formation energies and
        are ignored.
        Args:
            structure: pymatgen structure object
        """
        struct_species = structure.types_of_specie
        if len(struct_species) == 1:
            oxi_states = {struct_species[0].symbol: 0}
        else:
            vir = VIRE(structure)
            oxi_states = vir.valences
        self.oxi_states = {}
        for key, val in oxi_states.items():
            strip_key = "".join([s for s in key if s.isalpha()])
            self.oxi_states[strip_key] = val

        self.min_max_oxi = {}
        for s in struct_species:
            if isinstance(s, Specie):
                el = s.element
            elif isinstance(s, Element):
                el = s
            else:
                continue
            max_oxi = max(el.common_oxidation_states)
            min_oxi = min(el.common_oxidation_states)
            self.min_max_oxi[el.symbol] = (min_oxi, max_oxi)

    def get_charges(self, defect_type, site_specie=None, sub_specie=None):
        """
        Based on the type of defect, site and substitution (if any) species
        the defect charge states are generated.
        Args:
            defect_type (str): Options are vacancy, antisite, substitution,
                               and interstitial
            site_specie: Specie on the host lattice site
                         For interstitials, use this
            sub_specie: Specie that is replacing the site specie.
                        For antisites and substitution defects
        """
        if defect_type == "vacancy":
            vac_symbol = get_el_sp(site_specie).symbol
            vac_oxi_state = self.oxi_states[vac_symbol]
            if vac_oxi_state < 0:
                min_oxi = max(vac_oxi_state, self.min_max_oxi[vac_symbol][0])
                max_oxi = 0
            elif vac_oxi_state > 0:
                max_oxi = min(vac_oxi_state, self.min_max_oxi[vac_symbol][1])
                min_oxi = 0
            else:  # most probably single element
                oxi_states = get_el_sp(site_specie).common_oxidation_states
                min_oxi = min(oxi_states)
                max_oxi = max(oxi_states)
            return [-c for c in range(min_oxi, max_oxi + 1)]

        elif defect_type == "antisite":
            vac_symbol = get_el_sp(site_specie).symbol
            vac_oxi_state = self.oxi_states[vac_symbol]
            as_symbol = get_el_sp(sub_specie).symbol
            if vac_oxi_state > 0:
                oxi_max = max(self.min_max_oxi[as_symbol][1], 0)
                oxi_min = 0
            else:
                oxi_max = 0
                oxi_min = min(self.min_max_oxi[as_symbol][0], 0)
            return [c - vac_oxi_state for c in range(oxi_min, oxi_max + 1)]

        elif defect_type == "substitution":
            site_specie = get_el_sp(site_specie)
            sub_specie = get_el_sp(sub_specie)
            vac_symbol = site_specie.symbol
            vac_oxi_state = self.oxi_states[vac_symbol]

            max_oxi_sub = max(sub_specie.common_oxidation_states)
            min_oxi_sub = min(sub_specie.common_oxidation_states)
            if vac_oxi_state > 0:
                if max_oxi_sub < 0:
                    raise ValueError("Substitution seems not possible")
                else:
                    if max_oxi_sub > vac_oxi_state:
                        return list(range(max_oxi_sub - vac_oxi_state + 1))
                    else:
                        return [max_oxi_sub - vac_oxi_state]
            else:
                if min_oxi_sub > 0:
                    raise ValueError("Substitution seems not possible")
                else:
                    if min_oxi_sub < vac_oxi_state:
                        return list(range(min_oxi_sub - vac_oxi_state, 1))
                    else:
                        return [min_oxi_sub - vac_oxi_state]

        elif defect_type == "interstitial":
            site_specie = get_el_sp(site_specie)
            min_oxi = min(min(site_specie.common_oxidation_states), 0)
            max_oxi = max(max(site_specie.common_oxidation_states), 0)

            return list(range(min_oxi, max_oxi + 1))


class DefectChargerIonic(DefectCharger):
    """
    Charge assignments based on values expected purely from ionic theory, range to zero.
    Simple but good for first guesses.
    """

    def __init__(self, structure):
        """
        Args:
            structure: pymatgen structure object
        """
        struct_species = structure.types_of_specie
        if len(struct_species) == 1:
            oxi_states = {struct_species[0].symbol: 0}
        else:
            vir = VIRE(structure)
            oxi_states = vir.valences
        self.oxi_states = {}
        for key, val in oxi_states.items():
            strip_key = "".join([s for s in key if s.isalpha()])
            self.oxi_states[strip_key] = val

    def get_charges(self, defect_type, site_specie=None, sub_specie=None):
        """
        Based on the type of defect, site and substitution (if any) species
        the defect charge states are generated.
        Args:
            defect_type (str): Options are vacancy, antisite, substitution,
                               and interstitial
            site_specie: Specie on the host lattice site
                         For interstitials, use this
            sub_specie: Specie that is replacing the site specie.
                        For antisites and substitution defects
        """
        if defect_type == "vacancy":
            vac_symbol = get_el_sp(site_specie).symbol
            vac_oxi_state = self.oxi_states[vac_symbol]
            if vac_oxi_state == 0:
                return [-1, 0, 1]
            else:
                minval = min(-vac_oxi_state, 0)
                maxval = max(-vac_oxi_state, 0)
                return [c for c in range(minval - 1, maxval + 2)]

        elif defect_type in ["antisite", "substitution"]:
            # TODO: may cause some weird states for substitutions. Worth updating in future.
            vac_symbol = get_el_sp(site_specie).symbol
            vac_oxi_state = self.oxi_states[vac_symbol]
            as_symbol = get_el_sp(sub_specie).symbol
            as_oxi_state = self.oxi_states[as_symbol]
            expected_oxi = as_oxi_state - vac_oxi_state
            if expected_oxi == 0:
                return [-1, 0, 1]
            else:
                minval = min(expected_oxi, 0)
                maxval = max(expected_oxi, 0)
                return [c for c in range(minval - 1, maxval + 2)]

        elif defect_type == "interstitial":
            return [-1, 0, 1]


class DefectChargerUserCustom(DefectCharger):
    """
        NOTE from developers:
            This code is for full control over charging approach of defects
            but will be replaced with a future version.
            Code has no unit test and will not be maintained
            going forward (as of 12/15/2017).
            However, we are keeping function here to allow for
            current users to make use of it...

    Determine oxidation states from bond valence method
    (unless oxidation states specified)
    Then ask user what charges they want
    """

    def __init__(self, structure, oxi_states=None):
        """
        Does initial Valence bond method to initialize oxidation states for user help
        Overwridden by oxi_state specified by
        Args: structure
            pymatgen structure object to determine the oxidation states
            min_max_oxi: any user specified min/max oxidation ranges for elements in structure
            oxi_states: any user specified oxidation states of elements in structure
        """
        oxi_states = oxi_states if oxi_states is not None else {}
        struct_species = structure.types_of_specie
        if (len(struct_species) == 1) and struct_species[
            0
        ].symbol not in oxi_states.keys():
            oxi_states[struct_species[0].symbol] = 0
        else:
            vir = VIRE(structure)
            for elt, oxi in vir.valences.items():
                strip_key = "".join([s for s in elt if s.isalpha()])
                if strip_key not in oxi_states.keys():
                    oxi_states[strip_key] = oxi
        self.oxi_states = oxi_states
        print(
            "\nThis is Full-User Charge Generation Mode.\n"
            "Options are: (1) Range mode (input min and max of each each type of defect) "
            "or (2) Individual mode (input each charge state you want for each defect)\n"
            "\nWhen finished with specific defect, press ENTER to continue."
        )
        rng_mod = raw_input("Please specify Range (R) or Individual (I) Mode:")
        if "R" == rng_mod.upper()[0]:
            self.rangemode = True
        else:
            self.rangemode = False

    def get_charges(self, defect_type, site_specie=None, sub_specie=None):
        """
        Based on the type of defect, site and substitution (if any) species
        the defect charge states are generated.
        Args:
            defect_type (str): Options are vacancy, antisite, substitution,
                               and interstitial
            site_specie: Specie on the host lattice site
                         For interstitials, use this
            sub_specie: Specie that is replacing the site specie.
                        For antisites and substitution defects
        """
        if site_specie in self.oxi_states.keys():
            sitechg = self.oxi_states[site_specie]
        else:
            sitechg = False

        if sub_specie:
            if sub_specie in self.oxi_states.keys():
                subchg = self.oxi_states[sub_specie]
            else:
                subchg = False

        def get_users_charges():
            tmpchgs = raw_input("What charges would you like? : ")
            chgs = [int(c) for c in tmpchgs.split()]
            if self.rangemode:
                return list(range(chgs[0], chgs[-1] + 1))
            else:
                return list(chgs)

        if defect_type == "vacancy":
            if not sitechg:
                print(
                    site_specie,
                    defect_type,
                    "charge suggestion unknown (specify oxidation states to get suggestion)",
                )
            else:
                print(
                    site_specie,
                    defect_type,
                    "has charge =",
                    -sitechg,
                    "according to Simple Ionic Theory",
                )
        elif defect_type in ["antisite", "substitution"]:
            nom = sub_specie + "_on_" + site_specie
            if not sitechg or not subchg:
                print(
                    nom,
                    defect_type,
                    "charge suggestion unknown (specify oxidation states to get suggestion)",
                )
            else:
                print(
                    nom,
                    defect_type,
                    "has charge = ",
                    subchg - sitechg,
                    "according to Simple Ionic Theory",
                )
        elif defect_type == "interstitial":
            if not sitechg:
                print(
                    site_specie,
                    defect_type,
                    "charge suggestion unknown (specify oxidation states to get suggestion)",
                )
            else:
                print(
                    site_specie,
                    defect_type,
                    "has charge = ",
                    sitechg,
                    "according to Simple Ionic Theory",
                )
        outchgs = get_users_charges()
        print("    Charges generated:", outchgs)
        return outchgs


class ChargedDefectsStructures(object):
    """
    A class to generate charged defective structures for use in first
    principles supercell formalism. The standard defects such as antisites
    and vacancies are generated.  Interstitial finding is also implemented
    (optional).
    """

    def __init__(
        self,
        structure,
        max_min_oxi=None,
        substitutions=None,
        oxi_states=None,
        cellmax=128,
        antisites_flag=True,
        include_interstitials=True,
        interstitial_elements=None,
        intersites=None,
        standardized=False,
        struct_type="semiconductor",
    ):
        """
        Args:
            structure (Structure):
                the bulk structure.
            max_min_oxi (dict):
                The minimal and maximum oxidation state of each element as a
                dict. For instance {"O":(-2,0)}. If not given, the oxi-states
                of pymatgen are considered.
            substitutions (dict):
                The allowed substitutions of elements as a dict. If not given,
                intrinsic defects are computed. If given, intrinsic (e.g.,
                anti-sites) and extrinsic are considered explicitly specified.
                Example: {"Co":["Zn","Mn"]} means Co sites can be substituted
                by Mn or Zn.
            oxi_states (dict):
                The oxidation state of the elements in the compound e.g.
                {"Fe":2,"O":-2}. If not given, the oxidation state of each
                site is computed with bond valence sum. WARNING: Bond-valence
                method can fail for mixed-valence compounds.
            cellmax (int):
                Maximum number of atoms allowed in the supercell.
            antisites_flag (bool):
                If False, don't generate antisites.
            include_interstitials (bool):
                If true, do generate interstitial defect configurations
                (default: False).
            interstitial_elements ([str]):
                List of strings containing symbols of the elements that are
                to be considered for interstitial sites.  The default (None)
                triggers self-interstitial generation,
                given that include_interstitials is True.
            intersites ([PeriodicSite]):
                A list of PeriodicSites in the bulk structure on which we put
                interstitials.  Note that you still have to set flag
                include_interstitials to True in order to make use of this
                manual way of providing interstitial sites.
                If this is used, then no additional interstitials are generated
                beyond the list that is provided in intersites.
            standardized (bool):
                If True, use the primitive standard structure as unit cell
                for generating the defect configurations (default is False).
                The primitive standard structure is obtained from the
                SpacegroupAnalyzer class with a symprec of 0.01.
            struct_type (string):
                Options are 'semiconductor' and 'insulator'. If semiconductor
                is selected, charge states based on database of semiconductors
                is used to assign defect charges. For insulators, defect
                charges are conservatively assigned.
        """
        max_min_oxi = max_min_oxi if max_min_oxi is not None else {}
        substitutions = substitutions if substitutions is not None else {}
        oxi_states = oxi_states if oxi_states is not None else {}
        interstitial_elements = (
            interstitial_elements if interstitial_elements is not None else []
        )
        intersites = intersites if intersites is not None else []

        self.defects = []
        self.cellmax = cellmax
        self.substitutions = {}
        self.struct_type = struct_type
        for key, val in substitutions.items():
            self.substitutions[key] = val

        spa = SpacegroupAnalyzer(structure, symprec=1e-2)
        prim_struct = spa.get_primitive_standard_structure()
        if standardized:
            self.struct = prim_struct
        else:
            self.struct = structure

        struct_species = self.struct.types_of_specie
        if self.struct_type == "semiconductor":
            self.defect_charger = DefectChargerSemiconductor(
                self.struct, min_max_oxi=max_min_oxi
            )
        elif self.struct_type == "insulator":
            self.defect_charger = DefectChargerInsulator(self.struct)
        elif self.struct_type == "manual":
            self.defect_charger = DefectChargerUserCustom(
                self.struct, oxi_states=oxi_states
            )
        elif self.struct_type == "ionic":
            self.defect_charger = DefectChargerIonic(self.struct)
        else:
            raise NotImplementedError

        if include_interstitials and interstitial_elements:
            for elem_str in interstitial_elements:
                if not Element.is_valid_symbol(elem_str):
                    raise ValueError(f"invalid interstitial element {elem_str}")

        sc_scale = get_optimized_sc_scale(self.struct, cellmax)
        self.defects = {}
        sc = self.struct.copy()
        sc.make_supercell(sc_scale)
        self.defects["bulk"] = {
            "name": "bulk",
            "supercell": {"size": sc_scale, "structure": sc},
        }

        # If interstitials are provided as a list of PeriodicSites,
        # make sure that the lattice has not changed.
        if include_interstitials and intersites:
            for intersite in intersites:  # list of PeriodicSite objects
                if intersite.lattice != self.struct.lattice:
                    raise RuntimeError(
                        "Discrepancy between lattices"
                        " underlying the input interstitials and"
                        " the bulk structure; possibly because of"
                        " standardizing the input structure."
                    )

        vacancies = []
        as_defs = []
        sub_defs = []

        VG = VacancyGenerator(self.struct)
        print("Setting up vacancies")
        for i, vac in enumerate(VG):
            vac_site = vac.site
            vac_symbol = vac.site.specie.symbol
            vac_sc = vac.generate_defect_structure(sc_scale)

            # create a trivial defect structure to find where supercell transformation moves the lattice
            struct_for_defect_site = Structure(
                vac.bulk_structure.copy().lattice,
                [vac.site.specie],
                [vac.site.frac_coords],
                to_unit_cell=True,
                coords_are_cartesian=False,
            )
            struct_for_defect_site.make_supercell(sc_scale)
            vac_sc_site = struct_for_defect_site[0]

            charges_vac = self.defect_charger.get_charges("vacancy", vac_symbol)

            for c in SimpleChargeGenerator(vac):
                vacancies.append(
                    {
                        "name": f"vac_{i+1}_{vac_symbol}",
                        "unique_site": vac_site,
                        "bulk_supercell_site": vac_sc_site,
                        "defect_type": "vacancy",
                        "site_specie": vac_symbol,
                        "site_multiplicity": vac.multiplicity,
                        "supercell": {"size": sc_scale, "structure": vac_sc},
                        "charges": charges_vac,
                        "Possible_KV_Charge": c.charge,
                    }
                )

        if antisites_flag:
            print("Setting up antisites")
            for as_specie in set(struct_species):
                SG = SubstitutionGenerator(self.struct, as_specie)
                for i, sub in enumerate(SG):
                    as_symbol = as_specie.symbol
                    as_sc = sub.generate_defect_structure(sc_scale)

                    # create a trivial defect structure to find where supercell transformation
                    # moves the defect
                    struct_for_defect_site = Structure(
                        sub.bulk_structure.copy().lattice,
                        [sub.site.specie],
                        [sub.site.frac_coords],
                        to_unit_cell=True,
                        coords_are_cartesian=False,
                    )
                    struct_for_defect_site.make_supercell(sc_scale)
                    as_sc_site = struct_for_defect_site[0]

                    # get bulk_site (non sc)
                    poss_deflist = sorted(
                        sub.bulk_structure.get_sites_in_sphere(
                            sub.site.coords, 0.01, include_index=True
                        ),
                        key=lambda x: x[1],
                    )
                    if not len(poss_deflist):
                        raise ValueError(
                            f"Could not find substitution site inside bulk structure for {sub.name}?"
                        )
                    defindex = poss_deflist[0][2]
                    as_site = sub.bulk_structure[defindex]
                    vac_symbol = as_site.specie

                    charges_as = self.defect_charger.get_charges(
                        "antisite", vac_symbol, as_symbol
                    )

                    for c in SimpleChargeGenerator(sub):
                        as_defs.append(
                            {
                                "name": f"as_{i+1}_{as_symbol}_on_{vac_symbol}",
                                "unique_site": as_site,
                                "bulk_supercell_site": as_sc_site,
                                "defect_type": "antisite",
                                "site_specie": vac_symbol,
                                "substituting_specie": as_symbol,
                                "site_multiplicity": sub.multiplicity,
                                "supercell": {"size": sc_scale, "structure": as_sc},
                                "charges": charges_as,
                                "Possible_KV_Charge": c.charge,
                            }
                        )

        for vac_symbol, subspecie_list in self.substitutions.items():
            for subspecie_symbol in subspecie_list:
                SG = SubstitutionGenerator(self.struct, subspecie_symbol)
                for i, sub in enumerate(SG):
                    sub_symbol = sub.site.specie.symbol

                    # get bulk_site (non sc)
                    poss_deflist = sorted(
                        sub.bulk_structure.get_sites_in_sphere(
                            sub.site.coords, 0.1, include_index=True
                        ),
                        key=lambda x: x[1],
                    )
                    if not len(poss_deflist):
                        raise ValueError(
                            f"Could not find substitution site inside bulk structure for {sub.name}?"
                        )
                    defindex = poss_deflist[0][2]
                    sub_site = self.struct[defindex]
                    this_vac_symbol = sub_site.specie.symbol

                    if (sub_symbol != subspecie_symbol) or (
                        this_vac_symbol != vac_symbol
                    ):
                        continue
                    else:
                        sub_sc = sub.generate_defect_structure(sc_scale)

                        # create a trivial defect structure to find where supercell
                        # transformation moves the defect
                        struct_for_defect_site = Structure(
                            sub.bulk_structure.copy().lattice,
                            [sub.site.specie],
                            [sub.site.frac_coords],
                            to_unit_cell=True,
                            coords_are_cartesian=False,
                        )
                        struct_for_defect_site.make_supercell(sc_scale)
                        sub_sc_site = struct_for_defect_site[0]

                        charges_sub = self.defect_charger.get_charges(
                            "substitution", vac_symbol, subspecie_symbol
                        )
                        for c in SimpleChargeGenerator(sub):
                            sub_defs.append(
                                {
                                    "name": f"sub_{i+1}_{subspecie_symbol}_on_{vac_symbol}",
                                    "unique_site": sub_site,
                                    "bulk_supercell_site": sub_sc_site,
                                    "defect_type": "substitution",
                                    "site_specie": vac_symbol,
                                    "substitution_specie": subspecie_symbol,
                                    "site_multiplicity": sub.multiplicity,
                                    "supercell": {
                                        "size": sc_scale,
                                        "structure": sub_sc,
                                    },
                                    "charges": charges_sub,
                                    "Possible_KV_Charge": c.charge,
                                }
                            )

        self.defects["vacancies"] = vacancies
        self.defects["substitutions"] = sub_defs
        self.defects["substitutions"] += as_defs

        if include_interstitials:
            interstitials = []

            if interstitial_elements:
                inter_elems = interstitial_elements
            else:
                inter_elems = [elem.symbol for elem in self.struct.composition.elements]
            if len(inter_elems) == 0:
                raise RuntimeError("empty element list for interstitials")

            if intersites:
                print("Setting up interstitials from intersites")
                # manual specification of interstitials
                for i, intersite in enumerate(intersites):
                    elt = intersite.specie
                    name = f"inter_{i+1}_{elt}"

                    if intersite.lattice != self.struct.lattice:
                        err_msg = "Lattice matching error occurs between provided interstitial and the bulk structure."
                        if standardized:
                            err_msg += (
                                "\nLikely because the standardized flag was used. Turn this flag off or reset "
                                "your interstitial PeriodicSite to match the standardized form of the bulk structure."
                            )
                        raise ValueError(err_msg)
                    else:
                        intersite_object = Interstitial(self.struct, intersite)

                    # create a trivial defect structure to find where supercell transformation
                    # moves the defect site
                    struct_for_defect_site = Structure(
                        intersite_object.bulk_structure.copy().lattice,
                        [intersite_object.site.specie],
                        [intersite_object.site.frac_coords],
                        to_unit_cell=True,
                        coords_are_cartesian=False,
                    )
                    struct_for_defect_site.make_supercell(sc_scale)
                    site_sc = struct_for_defect_site[0]

                    sc_with_inter = intersite_object.generate_defect_structure(sc_scale)
                    charges_inter = self.defect_charger.get_charges("interstitial", elt)

                    for c in SimpleChargeGenerator(intersite_object):
                        interstitials.append(
                            {
                                "name": name,
                                "unique_site": intersite_object.site,
                                "bulk_supercell_site": site_sc,
                                "defect_type": "interstitial",
                                "site_specie": intersite_object.site.specie.symbol,
                                "site_multiplicity": intersite_object.multiplicity,
                                "supercell": {
                                    "size": sc_scale,
                                    "structure": sc_with_inter,
                                },
                                "charges": charges_inter,
                                "Possible_KV_Charge": c.charge,
                            }
                        )

            else:
                print(
                    "Searching for Voronoi interstitial sites (this can take a little while)"
                )
                # the use of O here is completely arbitrary
                IG = VoronoiInterstitialGenerator(self.struct, "O")
                sites = []
                for i in IG:
                    sites.append(i)

                inters = []
                for el in inter_elems:
                    for s in sites:
                        d = PeriodicSite(
                            el, s.site.frac_coords, s.bulk_structure.copy().lattice
                        )
                        inters.append(d)

                print("Found the interstital sites, setting up interstitials")
                for i, intersite in enumerate(inters):
                    elt = intersite.specie
                    if intersite.lattice != self.struct.lattice:
                        err_msg = (
                            "Lattice matching error occurs between provided interstitial "
                            "and the bulk structure."
                        )
                        raise ValueError(err_msg)
                    else:
                        intersite_object = Interstitial(self.struct, intersite)

                    name = f"inter_{i+1}_{elt}"

                    # create a trivial defect structure to find where supercell transformation
                    # moves the defect site
                    struct_for_defect_site = Structure(
                        intersite_object.bulk_structure.copy().lattice,
                        [intersite_object.site.specie],
                        [intersite_object.site.frac_coords],
                        to_unit_cell=True,
                        coords_are_cartesian=False,
                    )
                    struct_for_defect_site.make_supercell(sc_scale)
                    site_sc = struct_for_defect_site[0]

                    sc_with_inter = intersite_object.generate_defect_structure(sc_scale)
                    charges_inter = self.defect_charger.get_charges("interstitial", elt)

                    for c in SimpleChargeGenerator(intersite_object):
                        interstitials.append(
                            {
                                "name": name,
                                "unique_site": intersite_object.site,
                                "bulk_supercell_site": site_sc,
                                "defect_type": "interstitial",
                                "site_specie": intersite_object.site.specie.symbol,
                                "site_multiplicity": intersite_object.multiplicity,
                                "supercell": {
                                    "size": sc_scale,
                                    "structure": sc_with_inter,
                                },
                                "charges": charges_inter,
                                "Possible_KV_Charge": c.charge,
                            }
                        )

            self.defects["interstitials"] = interstitials

        print("Defects generated:")
        for key, defect_list in self.defects.items():
            if key != "bulk" and len(defect_list) > 0:
                table = []
                for single_defect_dict in defect_list:
                    header = [
                        key.capitalize(),
                        "Potential Charge States",
                        "Supercell Site " "Multiplicity",
                    ]
                    row = [
                        single_defect_dict["name"],
                        single_defect_dict["charges"],
                        single_defect_dict["site_multiplicity"],
                    ]
                    table.append(row)
                print(
                    tabulate(
                        table,
                        headers=header,
                        stralign="left",
                        numalign="left",
                    ),
                    "\n",
                )

    def to(self, outfile):
        dumpfn(self.defects, outfile)

    def get_n_defects_of_type(self, defect_type):
        """
        Get the number of defects of the given type.
        Args:
            defect_type (str): defect type (vacancies,
                interstitials, substitutions).
        Returns:
            n_defects (int): number of defects of given type.
        """
        try:
            return len(self.defects[defect_type])
        except:
            return 0

    def get_ith_supercell_of_defect_type(self, i, defect_type):
        """
        Get the (i-1)-th defect supercell with the given defect type.
        Args:
            i (int): index (starting from 0) of target supercell.
            defect_type (str): defect type (vacancies,
                interstitials, substitutions).
        Returns:
            sc (Structure): copy of the defect supercell.
        """
        return self.defects[defect_type][i]["supercell"]["structure"].copy()



import warnings
from datetime import date
from importlib.metadata import PackageNotFoundError, version

from pymatgen.io.vasp.inputs import UnknownPotcarWarning

if date.today().weekday() in [5, 6]:
    print("""Working on the weekend, like usual...\n""")
if date.today().weekday() == 5:
    print("Seriously though, everyone knows Saturday's for the boys/girls...\n")


def _check_pmg_compatibility():
    try:
        v_pmg_def = version("pymatgen-analysis-defects")
    except PackageNotFoundError:
        v_pmg_def = False

    try:
        v_pmg = version("pymatgen")
    except Exception:
        v_pmg = False

    if v_pmg_def:
        raise TypeError(
            "You have the `pymatgen-analysis-defects` package installed, which is currently "
            "incompatible with `doped`. Please uninstall `pymatgen-analysis-defects` (with `pip "
            "uninstall pymatgen-analysis-defects`) and restart the kernel."
        )

    if v_pmg > "2022.7.25":
        raise TypeError(
            f"You have the version {v_pmg} of `pymatgen` installed, which is currently "
            f"incompatible with `doped`. Please revert this package (with `pip install "
            f"pymatgen==2022.7.25`) and restart the kernel."
        )


_check_pmg_compatibility()

# globally ignore these POTCAR warnings
warnings.filterwarnings("ignore", category=UnknownPotcarWarning)
warnings.filterwarnings("ignore", message="No POTCAR file with matching TITEL fields")
warnings.filterwarnings("ignore", message="Ignoring unknown variable type")
warnings.filterwarnings(
        "ignore", message="POTCAR data with symbol"
    )  # Ignore POTCAR warnings because Pymatgen incorrectly detecting POTCAR types
# Ignore because comment after 'ALGO = Normal' causes this unnecessary warning:
warnings.filterwarnings("ignore", message="Hybrid functionals only support")

# until updated from pymatgen==2022.7.25 :
warnings.filterwarnings(
    "ignore", message="Using `tqdm.autonotebook.tqdm` in notebook mode"
)
warnings.filterwarnings("ignore", message="`np.int` is a deprecated alias for the builtin `int`")
warnings.filterwarnings("ignore", message="Use get_magnetic_symmetry()")



import os
import shutil
import unittest
from pathlib import Path

from monty.serialization import loadfn
from pymatgen.core.structure import Structure

from doped import chemical_potentials


class ChemPotsTestCase(unittest.TestCase):
    def setUp(self):
        self.path = Path(__file__).parents[1].joinpath("examples/competing_phases")
        self.stable_system = "ZrO2"
        self.unstable_system = "Zr2O"
        self.extrinsic_species = "La"
        self.csv_path = self.path / "zro2_competing_phase_energies.csv"
        self.csv_path_ext = self.path / "zro2_la_competing_phase_energies.csv"
        self.parsed_chempots = loadfn(self.path / "zro2_chempots.json")
        self.parsed_ext_chempots = loadfn(self.path / "zro2_la_chempots.json")

    def tearDown(self) -> None:
        if Path("chempot_limits.csv").is_file():
            os.remove("chempot_limits.csv")

        if Path("competing_phases.csv").is_file():
            os.remove("competing_phases.csv")

        if Path("input.dat").is_file():
            os.remove("input.dat")

        return super().tearDown()

    def test_cpa_csv(self):
        stable_cpa = chemical_potentials.CompetingPhasesAnalyzer(self.stable_system)
        stable_cpa.from_csv(self.csv_path)
        self.ext_cpa = chemical_potentials.CompetingPhasesAnalyzer(
            self.stable_system, self.extrinsic_species
        )
        self.ext_cpa.from_csv(self.csv_path_ext)

        self.assertEqual(len(stable_cpa.elemental), 2)
        self.assertEqual(len(self.ext_cpa.elemental), 3)
        self.assertTrue(any(entry["formula"] == "O2" for entry in stable_cpa.data))
        self.assertAlmostEqual(
            [
                entry["energy_per_fu"]
                for entry in self.ext_cpa.data
                if entry["formula"] == "La2Zr2O7"
            ][0],
            -119.619571095,
        )

    # test chempots
    def test_cpa_chempots(self):
        stable_cpa = chemical_potentials.CompetingPhasesAnalyzer(self.stable_system)
        stable_cpa.from_csv(self.csv_path)
        df1 = stable_cpa.calculate_chempots()
        self.assertEqual(list(df1["O"])[0], 0)
        # check if it's no longer Element
        self.assertEqual(
            type(list(stable_cpa.intrinsic_chem_limits["elemental_refs"].keys())[0]),
            str,
        )

        self.unstable_cpa = chemical_potentials.CompetingPhasesAnalyzer(
            self.unstable_system
        )
        self.unstable_cpa.from_csv(self.csv_path)
        with self.assertRaises(ValueError):
            self.unstable_cpa.calculate_chempots()

        self.ext_cpa = chemical_potentials.CompetingPhasesAnalyzer(
            self.stable_system, self.extrinsic_species
        )
        self.ext_cpa.from_csv(self.csv_path_ext)
        df = self.ext_cpa.calculate_chempots()
        self.assertEqual(list(df["La_limiting_phase"])[0], "La2Zr2O7")
        self.assertAlmostEqual(list(df["La"])[0], -9.46298748)

    def test_cpa_chem_limits(self):
        # test accessing cpa.chem_limits without previously calling cpa.calculate_chempots()
        stable_cpa = chemical_potentials.CompetingPhasesAnalyzer(self.stable_system)
        stable_cpa.from_csv(self.csv_path)
        self.assertDictEqual(stable_cpa.chem_limits, self.parsed_chempots)

        self.ext_cpa = chemical_potentials.CompetingPhasesAnalyzer(
            self.stable_system, self.extrinsic_species
        )
        self.ext_cpa.from_csv(self.csv_path_ext)
        self.assertDictEqual(
            self.ext_cpa.chem_limits["elemental_refs"],
            self.parsed_ext_chempots["elemental_refs"],
        )

    # test vaspruns
    def test_vaspruns(self):
        cpa = chemical_potentials.CompetingPhasesAnalyzer(self.stable_system)
        path = self.path / "ZrO2"
        cpa.from_vaspruns(path=path, folder="relax", csv_fname=self.csv_path)
        self.assertEqual(len(cpa.elemental), 2)
        self.assertEqual(cpa.data[0]["formula"], "O2")

        cpa_no = chemical_potentials.CompetingPhasesAnalyzer(self.stable_system)
        with self.assertRaises(FileNotFoundError):
            cpa_no.from_vaspruns(path="path")

        with self.assertRaises(ValueError):
            cpa_no.from_vaspruns(path=0)

        ext_cpa = chemical_potentials.CompetingPhasesAnalyzer(
            self.stable_system, self.extrinsic_species
        )
        path2 = self.path / "La_ZrO2"
        ext_cpa.from_vaspruns(path=path2, folder="relax", csv_fname=self.csv_path_ext)
        self.assertEqual(len(ext_cpa.elemental), 3)
        # sorted by num_species, then alphabetically, then by num_atoms_in_fu, then by
        # formation_energy
        self.assertEqual(
            [entry["formula"] for entry in ext_cpa.data],
            [
                "La",
                "O2",
                "Zr",
                "Zr",
                "La2O3",
                "Zr2O",
                "Zr3O",
                "Zr3O",
                "ZrO2",
                "ZrO2",
                "La2Zr2O7",
            ],
        )
        # spot check some values
        # [{'formula': 'La', 'kpoints': '10x10x3', 'energy_per_fu': -5.00458616,
        # 'energy_per_atom': -5.00458616, 'energy': -20.01834464, 'formation_energy': 0.0},
        # {'formula': 'O2', 'kpoints': '2x2x2', 'energy_per_fu': -14.01320413, 'energy_per_atom':
        # -7.006602065, 'energy': -14.01320413, 'formation_energy': 0.0}, {'formula': 'Zr',
        # 'kpoints': '9x9x5', 'energy_per_fu': -9.84367624, 'energy_per_atom': -9.84367624,
        # 'energy': -19.68735248, 'formation_energy': 0.0}, {'formula': 'Zr', 'kpoints': '6x6x6',
        # 'energy_per_fu': -9.818516226666667, 'energy_per_atom': -9.818516226666667, 'energy':
        # -58.91109736, 'formation_energy': 0.025160013333334064}, {'formula': 'La2O3',
        # 'kpoints': '3x3x3', 'energy_per_fu': -49.046189555, 'energy_per_atom': -9.809237911,
        # 'energy': -392.36951644, 'formation_energy': -18.01721104}, {'formula': 'Zr2O',
        # 'kpoints': '5x5x2', 'energy_per_fu': -32.42291351666667, 'energy_per_atom':
        # -10.807637838888889, 'energy': -194.5374811, 'formation_energy': -5.728958971666668},
        # {'formula': 'Zr3O', 'kpoints': '5x5x5', 'energy_per_fu': -42.524204305,
        # 'energy_per_atom': -10.63105107625, 'energy': -85.04840861, 'formation_energy':
        # -5.986573519999993}, {'formula': 'Zr3O', 'kpoints': '5x5x5', 'energy_per_fu':
        # -42.472744875, 'energy_per_atom': -10.61818621875, 'energy': -84.94548975,
        # 'formation_energy': -5.935114089999992}, {'formula': 'ZrO2', 'kpoints': '3x3x3',
        # 'energy_per_fu': -34.83230881, 'energy_per_atom': -11.610769603333333, 'energy':
        # -139.32923524, 'formation_energy': -10.975428440000002}, {'formula': 'ZrO2', 'kpoints':
        # '3x3x1', 'energy_per_fu': -34.807990365, 'energy_per_atom': -11.602663455, 'energy':
        # -278.46392292, 'formation_energy': -10.951109995000003}, {'formula': 'La2Zr2O7',
        # 'kpoints': '3x3x3', 'energy_per_fu': -119.619571095, 'energy_per_atom':
        # -10.874506463181818, 'energy': -239.23914219, 'formation_energy': -40.87683184}]
        self.assertAlmostEqual(ext_cpa.data[0]["energy_per_fu"], -5.00458616)
        self.assertAlmostEqual(ext_cpa.data[0]["energy_per_atom"], -5.00458616)
        self.assertAlmostEqual(ext_cpa.data[0]["energy"], -20.01834464)
        self.assertAlmostEqual(ext_cpa.data[0]["formation_energy"], 0.0)
        self.assertAlmostEqual(ext_cpa.data[-1]["energy_per_fu"], -119.619571095)
        self.assertAlmostEqual(ext_cpa.data[-1]["energy_per_atom"], -10.874506463181818)
        self.assertAlmostEqual(ext_cpa.data[-1]["energy"], -239.23914219)
        self.assertAlmostEqual(ext_cpa.data[-1]["formation_energy"], -40.87683184)
        self.assertAlmostEqual(ext_cpa.data[6]["energy_per_fu"], -42.524204305)
        self.assertAlmostEqual(ext_cpa.data[6]["energy_per_atom"], -10.63105107625)
        self.assertAlmostEqual(ext_cpa.data[6]["energy"], -85.04840861)
        self.assertAlmostEqual(ext_cpa.data[6]["formation_energy"], -5.986573519999993)

        # check if it works from a list
        all_paths = []
        for p in path.iterdir():
            if not p.name.startswith("."):
                pp = p / "relax" / "vasprun.xml"
                ppgz = p / "relax" / "vasprun.xml.gz"
                if pp.is_file():
                    all_paths.append(pp)
                elif ppgz.is_file():
                    all_paths.append(ppgz)
        lst_cpa = chemical_potentials.CompetingPhasesAnalyzer(self.stable_system)
        lst_cpa.from_vaspruns(path=all_paths)
        self.assertEqual(len(lst_cpa.elemental), 2)
        self.assertEqual(len(lst_cpa.vasprun_paths), 8)

        all_fols = []
        for p in path.iterdir():
            if not p.name.startswith("."):
                pp = p / "relax"
                all_fols.append(pp)
        lst_fols_cpa = chemical_potentials.CompetingPhasesAnalyzer(self.stable_system)
        lst_fols_cpa.from_vaspruns(path=all_fols)
        self.assertEqual(len(lst_fols_cpa.elemental), 2)

    def test_cplap_input(self):
        cpa = chemical_potentials.CompetingPhasesAnalyzer(self.stable_system)
        cpa.from_csv(self.csv_path)
        cpa.cplap_input(dependent_variable="O")

        self.assertTrue(Path("input.dat").is_file())

        with open("input.dat", "r") as file:
            contents = file.readlines()

        # assert these lines are in the file:
        for i in [
            "2  # number of elements in bulk\n",
            "1 Zr 2 O -10.975428440000002  # num_atoms, element, formation_energy (bulk)\n",
            "O  # dependent variable (element)\n",
            "2  # number of bordering phases\n",
            "1  # number of elements in phase:\n",
            "2 O 0.0  # num_atoms, element, formation_energy\n",
            "2  # number of elements in phase:\n",
            "3 Zr 1 O -5.986573519999993  # num_atoms, element, formation_energy\n",
        ]:
            self.assertIn(i, contents)

        self.assertNotIn(
            "1 Zr 2 O -10.951109995000005\n", contents
        )  # shouldn't be included as is a higher energy polymorph of the bulk phase
        self.assertFalse(any("2 Zr 1 O" in i for i in contents))  # non-bordering phase


class BoxedMoleculesTestCase(unittest.TestCase):
    def test_elements(self):
        s, f, m = chemical_potentials.make_molecule_in_a_box("O2")
        self.assertEqual(f, "O2")
        self.assertEqual(m, 2)
        self.assertEqual(type(s), Structure)

        with self.assertRaises(UnboundLocalError):
            chemical_potentials.make_molecule_in_a_box("R2")


class FormationEnergyTestCase(unittest.TestCase):
    def test_fe(self):
        elemental = {"O": -7.006602065, "Zr": -9.84367624}
        data = [
            {
                "formula": "O2",
                "energy_per_fu": -14.01320413,
                "energy_per_atom": -7.006602065,
                "energy": -14.01320413,
            },
            {
                "formula": "Zr",
                "energy_per_fu": -9.84367624,
                "energy_per_atom": -9.84367624,
                "energy": -19.68735248,
            },
            {
                "formula": "Zr3O",
                "energy_per_fu": -42.524204305,
                "energy_per_atom": -10.63105107625,
                "energy": -85.04840861,
            },
            {
                "formula": "ZrO2",
                "energy_per_fu": -34.5391058,
                "energy_per_atom": -11.5130352,
                "energy": -138.156423,
            },
            {
                "formula": "ZrO2",
                "energy_per_fu": -34.83230881,
                "energy_per_atom": -11.610769603333331,
                "energy": -139.32923524,
            },
            {
                "formula": "Zr2O",
                "energy_per_fu": -32.42291351666667,
                "energy_per_atom": -10.807637838888889,
                "energy": -194.5374811,
            },
        ]

        df = chemical_potentials._calculate_formation_energies(data, elemental)
        print(df)
        self.assertEqual(df["formula"][0], "O2")  # definite order
        self.assertEqual(df["formation_energy"][0], 0)
        self.assertEqual(df["formula"][1], "Zr")
        self.assertEqual(df["formation_energy"][1], 0)
        self.assertAlmostEqual(
            df["formation_energy"][4], -10.975428440000002
        )  # lowest energy ZrO2


class CombineExtrinsicTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.path = Path(__file__).parents[1].joinpath("examples/competing_phases")
        first = self.path / "zro2_la_chempots.json"
        second = self.path / "zro2_y_chempots.json"
        self.first = loadfn(first)
        self.second = loadfn(second)
        self.extrinsic_species = "Y"
        return super().setUp()

    def test_combine_extrinsic(self):
        d = chemical_potentials.combine_extrinsic(
            self.first, self.second, self.extrinsic_species
        )
        self.assertEqual(len(d["elemental_refs"].keys()), 4)
        facets = list(d["facets"].keys())
        self.assertEqual(facets[0].rsplit("-", 1)[1], "Y2Zr2O7")

    def test_combine_extrinsic_errors(self):
        d = {"a": 1}
        with self.assertRaises(KeyError):
            chemical_potentials.combine_extrinsic(
                d, self.second, self.extrinsic_species
            )

        with self.assertRaises(KeyError):
            chemical_potentials.combine_extrinsic(self.first, d, self.extrinsic_species)

        with self.assertRaises(ValueError):
            chemical_potentials.combine_extrinsic(self.first, self.second, "R")


class CompetingPhasesTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.path = Path(__file__).parents[0]
        self.api_key = "c2LiJRMiBeaN5iXsH"  # SK MP Imperial email A/C
        self.cp = chemical_potentials.CompetingPhases(
            "ZrO2", e_above_hull=0.03, api_key=self.api_key
        )

        return super().setUp()

    def tearDown(self) -> None:
        if Path("competing_phases").is_dir():
            shutil.rmtree("competing_phases")

        return super().tearDown()

    def test_init(self):
        self.assertEqual(len(self.cp.entries), 13)
        self.assertEqual(self.cp.entries[0].name, "O2")
        self.assertEqual(self.cp.entries[0].data["total_magnetization"], 2)
        self.assertEqual(self.cp.entries[0].data["e_above_hull"], 0)
        self.assertTrue(self.cp.entries[0].data["molecule"])
        self.assertAlmostEqual(
            self.cp.entries[0].data["energy_per_atom"], -4.94795546875
        )
        self.assertAlmostEqual(self.cp.entries[0].data["energy"], -9.8959109375)
        self.assertEqual(self.cp.entries[1].name, "Zr")
        self.assertAlmostEqual(
            self.cp.entries[1].data["total_magnetization"], 0, places=3
        )
        self.assertEqual(self.cp.entries[1].data["e_above_hull"], 0)
        self.assertFalse(self.cp.entries[1].data["molecule"])
        self.assertEqual(self.cp.entries[2].name, "Zr3O")
        self.assertEqual(self.cp.entries[2].data["e_above_hull"], 0)
        self.assertEqual(self.cp.entries[3].name, "ZrO2")
        self.assertEqual(self.cp.entries[3].data["e_above_hull"], 0)
        self.assertEqual(self.cp.entries[4].name, "Zr3O")
        self.assertEqual(self.cp.entries[5].name, "Zr3O")
        self.assertEqual(self.cp.entries[6].name, "Zr2O")
        self.assertEqual(self.cp.entries[7].name, "ZrO2")
        self.assertEqual(self.cp.entries[8].name, "ZrO2")
        self.assertEqual(self.cp.entries[9].name, "Zr")
        self.assertEqual(self.cp.entries[10].name, "ZrO2")
        self.assertEqual(self.cp.entries[11].name, "ZrO2")
        self.assertEqual(self.cp.entries[12].name, "ZrO2")

        self.assertNotIn("Zr4O", [e.name for e in self.cp.entries])

    def test_init_full_phase_diagram(self):
        cp = chemical_potentials.CompetingPhases(
            "ZrO2", e_above_hull=0.03, api_key=self.api_key, full_phase_diagram=True
        )
        self.assertEqual(len(cp.entries), 14)  # Zr4O now present
        self.assertEqual(cp.entries[0].name, "O2")
        self.assertEqual(cp.entries[0].data["total_magnetization"], 2)
        self.assertEqual(cp.entries[0].data["e_above_hull"], 0)
        self.assertTrue(cp.entries[0].data["molecule"])
        self.assertAlmostEqual(cp.entries[0].data["energy_per_atom"], -4.94795546875)
        self.assertAlmostEqual(cp.entries[0].data["energy"], -9.8959109375)
        self.assertEqual(cp.entries[1].name, "Zr")
        self.assertAlmostEqual(cp.entries[1].data["total_magnetization"], 0, places=3)
        self.assertEqual(cp.entries[1].data["e_above_hull"], 0)
        self.assertFalse(cp.entries[1].data["molecule"])
        self.assertEqual(cp.entries[2].name, "Zr3O")
        self.assertEqual(cp.entries[2].data["e_above_hull"], 0)
        self.assertEqual(cp.entries[3].name, "Zr4O")  # new entry!
        self.assertEqual(cp.entries[3].data["e_above_hull"], 0)
        self.assertEqual(cp.entries[4].name, "ZrO2")
        self.assertEqual(cp.entries[4].data["e_above_hull"], 0)
        self.assertEqual(cp.entries[5].name, "Zr3O")
        self.assertEqual(cp.entries[6].name, "Zr3O")
        self.assertEqual(cp.entries[7].name, "Zr2O")
        self.assertEqual(cp.entries[8].name, "ZrO2")
        self.assertEqual(cp.entries[9].name, "ZrO2")
        self.assertEqual(cp.entries[10].name, "Zr")
        self.assertEqual(cp.entries[11].name, "ZrO2")
        self.assertEqual(cp.entries[12].name, "ZrO2")
        self.assertEqual(cp.entries[13].name, "ZrO2")

    def test_init_ytos(self):
        # 144 phases on Y-Ti-O-S MP phase diagram
        cp = chemical_potentials.CompetingPhases(
            "Y2Ti2S2O5", e_above_hull=0.1, api_key=self.api_key
        )
        self.assertEqual(len(cp.entries), 115)  # 115 phases with default algorithm
        # assert only one O2 phase present (molecular entry):
        o2_entries = [e for e in cp.entries if e.name == "O2"]
        self.assertEqual(len(o2_entries), 1)
        self.assertEqual(o2_entries[0].name, "O2")
        self.assertEqual(o2_entries[0].data["total_magnetization"], 2)
        self.assertEqual(o2_entries[0].data["e_above_hull"], 0)
        self.assertTrue(o2_entries[0].data["molecule"])
        self.assertAlmostEqual(o2_entries[0].data["energy_per_atom"], -4.94795546875)

        cp = chemical_potentials.CompetingPhases(
            "Y2Ti2S2O5", e_above_hull=0.1, full_phase_diagram=True, api_key=self.api_key
        )
        self.assertEqual(
            len(cp.entries), 140
        )  # 144 phases on Y-Ti-O-S MP phase diagram,
        # 4 extra O2 phases removed

        # assert only one O2 phase present (molecular entry):
        o2_entries = [e for e in cp.entries if e.name == "O2"]
        self.assertEqual(len(o2_entries), 1)
        self.assertEqual(o2_entries[0].name, "O2")
        self.assertEqual(o2_entries[0].data["total_magnetization"], 2)
        self.assertEqual(o2_entries[0].data["e_above_hull"], 0)
        self.assertTrue(o2_entries[0].data["molecule"])
        self.assertAlmostEqual(o2_entries[0].data["energy_per_atom"], -4.94795546875)

    def test_api_keys_errors(self):
        with self.assertRaises(ValueError) as e:
            nonvalid_api_key_error = ValueError(
                "API key test is not a valid legacy Materials Project API key. These are "
                "available at https://legacy.materialsproject.org/open"
            )
            chemical_potentials.CompetingPhases(
                "ZrO2",
                api_key="test",
            )
            self.assertIn(nonvalid_api_key_error, e.exception)

        with self.assertRaises(ValueError) as e:
            new_api_key_error = ValueError(
                "You are trying to use the new Materials Project (MP) API which is not supported "
                "by doped. Please use the legacy MP API (https://legacy.materialsproject.org/open)."
            )
            chemical_potentials.CompetingPhases(
                "ZrO2",
                api_key="testabcdefghijklmnopqrstuvwxyz12",
            )
            self.assertIn(new_api_key_error, e.exception)

    def test_convergence_setup(self):
        # potcar spec doesnt need potcars set up for pmg and it still works
        self.cp.convergence_setup(potcar_spec=True)
        self.assertEqual(len(self.cp.metals), 6)
        self.assertEqual(self.cp.metals[0].data["band_gap"], 0)
        self.assertFalse(self.cp.nonmetals[0].data["molecule"])
        # this shouldnt exist - dont need to convergence test for molecules
        self.assertFalse(Path("competing_phases/O2_EaH_0").is_dir())

        # test if it writes out the files correctly
        path1 = "competing_phases/ZrO2_EaH_0.0088/kpoint_converge/k2,1,1/"
        self.assertTrue(Path(path1).is_dir())
        with open(f"{path1}/KPOINTS", "r") as file:
            contents = file.readlines()
            self.assertEqual(contents[3], "2 1 1\n")

        with open(f"{path1}/POTCAR.spec", "r") as file:
            contents = file.readlines()
            self.assertEqual(contents[0], "Zr_sv\n")

        with open(f"{path1}/INCAR", "r") as file:
            contents = file.readlines()
            self.assertTrue(any("GGA = Ps\n" == line for line in contents))
            self.assertTrue(any("NSW = 0\n" == line for line in contents))

    def test_vasp_std_setup(self):
        self.cp.vasp_std_setup(potcar_spec=True)
        self.assertEqual(len(self.cp.nonmetals), 6)
        self.assertEqual(len(self.cp.metals), 6)
        self.assertEqual(len(self.cp.molecules), 1)
        self.assertEqual(self.cp.molecules[0].name, "O2")
        self.assertEqual(self.cp.molecules[0].data["total_magnetization"], 2)
        self.assertTrue(self.cp.molecules[0].data["molecule"])
        self.assertFalse(self.cp.nonmetals[0].data["molecule"])

        path1 = "competing_phases/ZrO2_EaH_0/vasp_std/"
        self.assertTrue(Path(path1).is_dir())
        with open(f"{path1}/KPOINTS", "r") as file:
            contents = file.readlines()
            self.assertEqual(
                contents[0], "pymatgen with grid density = 911 / number of atoms\n"
            )
            self.assertEqual(contents[3], "4 4 4\n")

        with open(f"{path1}/POTCAR.spec", "r") as file:
            contents = file.readlines()
            self.assertEqual(contents, ["Zr_sv\n", "O"])

        with open(f"{path1}/INCAR", "r") as file:
            contents = file.readlines()
            self.assertEqual(contents[0], "AEXX = 0.25\n")
            self.assertEqual(contents[8], "ISIF = 3\n")
            self.assertEqual(contents[5], "GGA = Pe\n")

        path2 = "competing_phases/O2_EaH_0/vasp_std"
        self.assertTrue(Path(path2).is_dir())
        with open(f"{path2}/KPOINTS", "r") as file:
            contents = file.readlines()
            self.assertEqual(contents[3], "1 1 1\n")

        with open(f"{path2}/POSCAR", "r") as file:
            contents = file.readlines()
            self.assertEqual(contents[-1], "0.500000 0.500000 0.540667 O\n")


class ExtrinsicCompetingPhasesTestCase(unittest.TestCase):
    # TODO: Need to add tests for co-doping, full_sub_approach, full_phase_diagram etc!!
    def setUp(self) -> None:
        self.path = Path(__file__).parents[0]
        self.api_key = "c2LiJRMiBeaN5iXsH"  # SK MP Imperial email A/C
        self.ex_cp = chemical_potentials.ExtrinsicCompetingPhases(
            "ZrO2", extrinsic_species="La", e_above_hull=0, api_key=self.api_key
        )
        return super().setUp()

    def tearDown(self) -> None:
        if Path("competing_phases").is_dir():
            shutil.rmtree("competing_phases")

        return super().tearDown()

    def test_init(self):
        self.assertEqual(len(self.ex_cp.entries), 2)
        self.assertEqual(self.ex_cp.entries[0].name, "La")  # definite ordering
        self.assertEqual(self.ex_cp.entries[1].name, "La2Zr2O7")  # definite ordering
        self.assertTrue(
            [entry.data["e_above_hull"] == 0 for entry in self.ex_cp.entries]
        )

        # names of intrinsic entries: ['O2', 'Zr', 'Zr3O', 'ZrO2']
        self.assertEqual(len(self.ex_cp.intrinsic_entries), 4)
        self.assertEqual(self.ex_cp.intrinsic_entries[0].name, "O2")
        self.assertEqual(self.ex_cp.intrinsic_entries[1].name, "Zr")
        self.assertEqual(self.ex_cp.intrinsic_entries[2].name, "Zr3O")
        self.assertEqual(self.ex_cp.intrinsic_entries[3].name, "ZrO2")
        self.assertTrue(
            all(
                entry.data["e_above_hull"] == 0
                for entry in self.ex_cp.intrinsic_entries
            )
        )

        cp = chemical_potentials.ExtrinsicCompetingPhases(
            "ZrO2", extrinsic_species="La", api_key=self.api_key
        )  # default e_above_hull=0.1
        self.assertEqual(len(cp.entries), 5)
        self.assertEqual(
            cp.entries[2].name, "La"
        )  # definite ordering, same 1st & 2nd as before
        self.assertEqual(cp.entries[3].name, "LaZr9O20")  # definite ordering
        self.assertEqual(cp.entries[4].name, "LaZr9O20")  # definite ordering
        self.assertTrue(
            all(entry.data["e_above_hull"] == 0 for entry in cp.entries[:2])
        )
        self.assertFalse(
            any(entry.data["e_above_hull"] == 0 for entry in cp.entries[2:])
        )
        self.assertEqual(len(cp.intrinsic_entries), 28)



# coding: utf-8

"""
Code to analyse VASP defect calculations.
These functions are built from a combination of useful modules from pymatgen, alongside
substantial modification, in the efforts of making an efficient, user-friendly package for
managing and analysing defect calculations, with publication-quality outputs
"""

import copy
from operator import itemgetter

import pandas as pd
from pymatgen.analysis.defects.thermodynamics import DefectPhaseDiagram
from pymatgen.util.string import unicodeify
from tabulate import tabulate

from doped import aide_murphy_correction


def bold_print(string: str) -> None:
    """Does what it says on the tin. Prints the input string in bold."""
    print("\033[1m" + string + "\033[0m")


def dpd_from_parsed_defect_dict(parsed_defect_dict: dict) -> DefectPhaseDiagram:
    """Generates a DefectPhaseDiagram object from a dictionary of parsed defect calculations (
    format: {"defect_name": defect_entry}), likely created using SingleDefectParser from
    doped.pycdt.utils.parse_calculations), which can then be used to analyse and plot the defect
    thermodynamics (formation energies, transition levels, concentrations etc)

    Args:
        parsed_defect_dict (dict):
            Dictionary of parsed defect calculations (format: {"defect_name": defect_entry}),
            likely created using SingleDefectParser from doped.pycdt.utils.parse_calculations).
            Must have 'vbm' and 'gap' in defect_entry.parameters for each defect (from
            SingleDefectParser.get_bulk_gap_data())

    Returns:
        pymatgen DefectPhaseDiagram object (DefectPhaseDiagram)
    """
    vbm_vals = []
    bandgap_vals = []
    for defect in parsed_defect_dict.values():
        vbm_vals.append(defect.parameters["vbm"])
        bandgap_vals.append(defect.parameters["gap"])
    if len(set(vbm_vals)) > 1:  # Check if all defects give same vbm
        raise ValueError(
            f"VBM values don't match for defects in given defect dictionary, "
            f"the VBM values in the dictionary are: {vbm_vals}. "
            f"Are you sure the correct/same bulk files were used with "
            f"SingleDefectParser and/or get_bulk_gap_data()?"
        )
    if len(set(bandgap_vals)) > 1:  # Check if all defects give same bandgap
        raise ValueError(
            f"Bandgap values don't match for defects in given defect dictionary, "
            f"the bandgap values in the dictionary are: {bandgap_vals}. "
            f"Are you sure the correct/same bulk files were used with "
            f"SingleDefectParser and/or get_bulk_gap_data()?"
        )
    vbm = vbm_vals[0]
    bandgap = bandgap_vals[0]
    dpd = DefectPhaseDiagram(
        list(parsed_defect_dict.values()), vbm, bandgap, filter_compatible=False
    )

    return dpd


def dpd_transition_levels(defect_phase_diagram: DefectPhaseDiagram):
    """Iteratively prints the charge transition levels for the input DefectPhaseDiagram object
    (via the from a defect_phase_diagram.transition_level_map attribute)

    Args:
        defect_phase_diagram (DefectPhaseDiagram):
            DefectPhaseDiagram object (likely created from
            analysis.dpd_from_parsed_defect_dict)

    Returns:
        None
    """
    for def_type, tl_info in defect_phase_diagram.transition_level_map.items():
        bold_print(f"\nDefect: {def_type.split('@')[0]}")
        for tl_efermi, chargeset in tl_info.items():
            print(
                f"Transition Level ({max(chargeset):{'+' if max(chargeset) else ''}}/"
                f"{min(chargeset):{'+' if min(chargeset) else ''}}) at {tl_efermi:.3f}"
                f" eV above the VBM"
            )


def formation_energy_table(
    defect_phase_diagram: DefectPhaseDiagram,
    chempot_limits: dict = None,
    pd_facets: list = None,
    fermi_level: float = 0,
    hide_cols: list = None,
    show_key: bool = True,
):
    """
    Prints defect formation energy tables for either a single chemical potential limit (i.e. phase
    diagram facet) or each facet in the phase diagram (chempot_limits dict), depending on the
    chempot_limits input supplied. This can either be a dictionary of chosen absolute/DFT chemical
    potentials: {Elt: Energy} (giving a single formation energy table) or a dictionary including
    the key-value pair: {"facets": [{'facet': [chempot_dict]}]}, following the format generated
    by chempot_limits = cpa.read_phase_diagram_and_chempots() (see example notebooks). In the
    latter case, a subset of facet(s) / chemical potential limit(s) can be chosen with the
    pd_facets argument, or if not specified, will print formation energy tables for each facet in
    the phase diagram.
    Returns the results a pandas DataFrame or list of DataFrames.

    Args:
        defect_phase_diagram (DefectPhaseDiagram):
             DefectPhaseDiagram object (likely created from
             analysis.dpd_from_parsed_defect_dict)
        chempot_limits (dict):
            This can either be a dictionary of chosen absolute/DFT chemical potentials: {Elt:
            Energy} (giving a single formation energy table) or a dictionary including the
            key-value pair: {"facets": [{'facet': [chempot_dict]}]}, following the format generated
            by chempot_limits = cpa.read_phase_diagram_and_chempots() (see example notebooks). If
            not specified, chemical potentials are not included in the formation energy calculation
            (all set to zero energy).
        pd_facets (list):
            A list facet(s) / chemical potential limit(s) for which to print the defect formation
            energy tables. If not specified, will print formation energy tables for each facet in
            the phase diagram. (default: None)
        fermi_level (float):
            Fermi level to use for computing the defect formation energies. (default: 0 (i.e.
            at the VBM))
        hide_cols: (list):
            List of columns to hide from the output. (default: None)
        show_key (bool):
            Whether or not to print the table key at the bottom of the output. (default: True)

    Returns:
        pandas DataFrame or list of DataFrames
    """
    if chempot_limits is None:
        chempot_limits = {}

    if "facets" in chempot_limits:
        list_of_dfs = []
        if not pd_facets:
            pd_facets = chempot_limits[
                "facets"
            ].keys()  # Phase diagram facets to use for chemical
            # potentials, to tabulate formation energies
        for facet in pd_facets:
            bold_print("Facet: " + unicodeify(facet))
            df = single_formation_energy_table(
                defect_phase_diagram,
                chempots=chempot_limits["facets"][facet],
                fermi_level=fermi_level,
                hide_cols=hide_cols,
                show_key=show_key,
            )
            list_of_dfs.append(df)
            print("\n")

        return list_of_dfs

    # else return {Elt: Energy} dict for chempot_limits, or if unspecified, all zero energy
    df = single_formation_energy_table(
        defect_phase_diagram,
        chempots=chempot_limits,
        fermi_level=fermi_level,
        hide_cols=hide_cols,
        show_key=show_key,
    )
    return df


def single_formation_energy_table(
    defect_phase_diagram: DefectPhaseDiagram,
    chempots: dict = None,
    fermi_level: float = 0,
    hide_cols: list = None,
    show_key: bool = True,
):
    """
    Prints a defect formation energy table for a single chemical potential limit (i.e. phase diagram
    facet), and returns the results as a pandas DataFrame.

    Args:
        defect_phase_diagram (DefectPhaseDiagram):
             DefectPhaseDiagram object (likely created from
             analysis.dpd_from_parsed_defect_dict)
        chempots (dict):
            Dictionary of chosen absolute/DFT chemical potentials: {Elt: Energy}. If not
            specified, chemical potentials are not included in the formation energy calculation
            (all set to zero energy).
        fermi_level (float):
            Fermi level to use for computing the defect formation energies. (default: 0 (i.e.
            at the VBM))
        hide_cols: (list):
            List of columns to hide from the output. (default: None)
        show_key (bool):
            Whether or not to print the table key at the bottom of the output. (default: True)

    Returns:
        pandas DataFrame sorted by formation energy
    """
    header = ["Defect", "Charge", "Defect Path"]
    table = []
    if hide_cols is None:
        hide_cols = []

    for defect_entry in defect_phase_diagram.entries:
        row = [
            defect_entry.name,
            defect_entry.charge,
            defect_entry.parameters["defect_path"],
        ]
        if "Uncorrected Energy" not in hide_cols:
            header += ["Uncorrected Energy"]
            row += [f"{defect_entry.uncorrected_energy:.2f} eV"]
        if "Corrected Energy" not in hide_cols:
            header += ["Corrected Energy"]
            row += [
                f"{defect_entry.energy:.2f} eV"
            ]  # With 0 chemical potentials, at the calculation
            # fermi level
        header += ["Formation Energy"]
        formation_energy = defect_entry.formation_energy(
            chemical_potentials=chempots, fermi_level=fermi_level
        )
        row += [f"{formation_energy:.2f} eV"]

        table.append(row)
    table = sorted(table, key=itemgetter(0, 1))
    print(
        tabulate(
            table,
            headers=header,
            tablefmt="fancy_grid",
            stralign="left",
            numalign="left",
        ),
        "\n",
    )

    if show_key:
        bold_print("Table Key:")
        print(
            """'Defect' -> Defect Type and Multiplicity
'Charge' -> Defect Charge State
'Uncorrected Energy' -> Defect Energy from calculation, without corrections
'Corrected Energy' -> Defect Energy from calculation (E_defect - E_host + corrections)
(chemical potentials set to 0 and the fermi level at average electrostatic potential in the
supercell)
'Formation Energy' -> Final Defect Formation Energy, with the specified chemical potentials (
chempot_limits)(default: all 0) and the chosen fermi_level (default: 0)(i.e. at the VBM)
        """
        )

    sorted_df = pd.DataFrame(
        table,
        columns=[
            "Defect",
            "Charge",
            "Defect Path",
            "Uncorrected Energy",
            "Corrected Energy",
            "Formation Energy",
        ],
    )
    sorted_df = sorted_df.sort_values("Formation Energy")
    return sorted_df


def lany_zunger_corrected_defect_dict_from_freysoldt(defect_dict: dict):
    """Convert input parsed defect dictionary (presumably created using SingleDefectParser
    from doped.pycdt.utils.parse_calculations) with Freysoldt charge corrections to
    the same parsed defect dictionary but with the Lany-Zunger charge correction (same potential
    alignment plus 0.65 * Makov-Payne image charge correction (same image charge correction as
    Freysoldt scheme)).
    Args:
        parsed_defect_dict (dict):
            Dictionary of parsed defect calculations (presumably created using SingleDefectParser
            from doped.pycdt.utils.parse_calculations) (see example notebook)
            Must have 'freysoldt_meta' in defect.parameters for each charged defect (from
            SingleDefectParser.freysoldt_loader())
    Returns:
        Parsed defect dictionary with Lany-Zunger charge corrections.
    """
    random_defect_entry = list(defect_dict.values())[
        0
    ]  # Just need any DefectEntry from
    # defect_dict to get the lattice and dielectric matrix
    lattice = random_defect_entry.bulk_structure.lattice.matrix
    dielectric = random_defect_entry.parameters["dielectric"]
    lz_image_charge_corrections = aide_murphy_correction.get_image_charge_correction(
        lattice, dielectric
    )
    lz_corrected_defect_dict = copy.deepcopy(defect_dict)
    for defect_name, defect_entry in lz_corrected_defect_dict.items():
        if defect_entry.charge != 0:
            potalign = defect_entry.parameters["freysoldt_meta"][
                "freysoldt_potential_alignment_correction"
            ]
            mp_pc_corr = lz_image_charge_corrections[
                abs(defect_entry.charge)
            ]  # Makov-Payne PC correction
            defect_entry.parameters.update(
                {
                    "Lany-Zunger_Corrections": {
                        "(Freysoldt)_Potential_Alignment_Correction": potalign,
                        "Makov-Payne_Image_Charge_Correction": mp_pc_corr,
                        "Lany-Zunger_Scaled_Image_Charge_Correction": 0.65 * mp_pc_corr,
                        "Total_Lany-Zunger_Correction": potalign + 0.65 * mp_pc_corr,
                    }
                }
            )
            defect_entry.corrections["charge_correction"] = defect_entry.parameters[
                "Lany-Zunger_Corrections"
            ]["Total_Lany-Zunger_Correction"]

        lz_corrected_defect_dict.update({defect_name: defect_entry})
    return lz_corrected_defect_dict


def lany_zunger_corrected_defect_dict_from_kumagai(defect_dict: dict):
    """Convert input parsed defect dictionary (presumably created using SingleDefectParser
    from doped.pycdt.utils.parse_calculations) with Kumagai charge corrections to
    the same parsed defect dictionary but with the 'Lany-Zunger' charge correction (same potential
    alignment plus 0.65 * image charge correction.
    Args:
        parsed_defect_dict (dict):
            Dictionary of parsed defect calculations (presumably created using SingleDefectParser
            from doped.pycdt.utils.parse_calculations) (see example notebook)
            Must have 'kumagai_meta' in defect.parameters for each charged defect (from
            SingleDefectParser.kumagai_loader())
    Returns:
        Parsed defect dictionary with Lany-Zunger charge corrections.
    """
    random_defect_entry = list(defect_dict.values())[
        0
    ]  # Just need any DefectEntry from
    # defect_dict to get the lattice and dielectric matrix
    lattice = random_defect_entry.bulk_structure.lattice.matrix
    dielectric = random_defect_entry.parameters["dielectric"]
    lz_image_charge_corrections = aide_murphy_correction.get_image_charge_correction(
        lattice, dielectric
    )
    lz_corrected_defect_dict = copy.deepcopy(defect_dict)
    for defect_name, defect_entry in lz_corrected_defect_dict.items():
        if defect_entry.charge != 0:
            potalign = defect_entry.parameters["kumagai_meta"][
                "kumagai_potential_alignment_correction"
            ]
            makove_payne_pc_correction = lz_image_charge_corrections[
                abs(defect_entry.charge)
            ]
            defect_entry.parameters.update(
                {
                    "Lany-Zunger_Corrections": {
                        "(Kumagai)_Potential_Alignment_Correction": potalign,
                        "Makov-Payne_Image_Charge_Correction": makove_payne_pc_correction,
                        "Lany-Zunger_Scaled_Image_Charge_Correction": 0.65
                        * makove_payne_pc_correction,
                        "Total_Lany-Zunger_Correction": potalign
                        + 0.65 * makove_payne_pc_correction,
                    }
                }
            )
            defect_entry.corrections["charge_correction"] = defect_entry.parameters[
                "Lany-Zunger_Corrections"
            ]["Total_Lany-Zunger_Correction"]

        lz_corrected_defect_dict.update({defect_name: defect_entry})
    return lz_corrected_defect_dict



# coding: utf-8

from __future__ import division

__status__ = "Development"

import os
import unittest

from doped.pycdt.corrections.ldau_correction import *

test_file_loc = os.path.abspath(
    os.path.join(__file__, "..", "..", "..", "..", "test_files")
)


class LDAUCorrectionTest(unittest.TestCase):
    def setUp(self):
        exp_gap, ggau_gap, gga_gap = (3, 2, 1)
        self.corrector = LDAUCorrection(exp_gap, ggau_gap, gga_gap)

    def test_transition_correction_with_identical_levels(self):
        self.assertAlmostEqual(0, self.corrector.get_transition_correction(0, 0))
        self.assertAlmostEqual(0, self.corrector.get_transition_correction(0.1, 0.1))

    def test_transition_correction_with_greater_lda_transition(self):
        self.assertAlmostEqual(-0.1, self.corrector.get_transition_correction(0.1, 0.2))

    def test_transition_correction_with_greater_ldau_transition(self):
        self.assertAlmostEqual(0.1, self.corrector.get_transition_correction(0.2, 0.1))

    def test_transition_correction_with_lda_transition_greater_than_bg(self):
        self.assertAlmostEqual(-1.1, self.corrector.get_transition_correction(0.4, 1.5))


import unittest

if __name__ == "__main__":
    unittest.main()
